#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
This example contain a ring resonator with input/output coupling waveguides.
"""

import numpy as np
from tdyno import TDyno

# ============   set up simulation   ==============
# initiate a simulation
tdyno = TDyno()

# geometry of the solving space
dx = 0.05
xmin = 0.
xmax = 20.5
dy = 0.05
ymin = 0.
ymax = 11.

# background permittivity and permeability. They fill the space.
epsi_bg = 1.
mu_bg = 1.

# choose polarization
polarization = 'Ez'  # either 'Hz' or 'Ez'

# time step, if simulation diverges, change into a smaller number
dt = 1. / 2. * np.sqrt(dx ** 2 + dy ** 2)

# Total time steps for each simulation.
# After it stops, you can click "play/pause" to continue for another Nt.
Nt = 20000

# solving mode
mode = 'HDE'  # for non-magnetic (permeability is 1 throughout), use 'HDE', otherwise, use 'BHDE'

# Initialize the solver
tdyno.setup(xmin, xmax, dx, ymin, ymax, dy, epsi_bg, mu_bg, polarization=polarization, dt=dt, Nt=Nt, mode=mode)  # this initializes the structure

# ============  add structures   ============
# waveguide 1
tdyno.add_structure(shape='rectangle',
                    xmin=0., xmax=1., ymin=1., ymax=1.5,
                    kind='dielectric',
                    epsi=12.)

# waveguide 2
tdyno.add_structure(shape='rectangle',
                    xmin=0., xmax=1., ymin=ymax - 1.5, ymax=ymax - 1.,
                    kind='dielectric',
                    epsi=12.)

# half circle wedge on left
tdyno.add_structure(shape='wedge',
                    xcenter=1., ycenter=5.5, radius=4.5, angle1=-90., angle2=90., width=0.5,
                    kind='dielectric',
                    epsi=12.)

# center ring
tdyno.add_structure(shape='ring',
                    xcenter=10.25, ycenter=5.5, radius=4.5, width=0.5,
                    kind='dielectric',
                    epsi=12.)

# right waveguide 1
tdyno.add_structure(shape='rectangle',
                    xmin=xmax - 1., xmax=xmax, ymin=1., ymax=1.5,
                    kind='dielectric',
                    epsi=12.)

# right waveguide 2
tdyno.add_structure(shape='rectangle',
                    xmin=xmax - 1., xmax=xmax, ymin=ymax - 1.5, ymax=ymax - 1.,
                    kind='dielectric',
                    epsi=12.)

# half circle wedge on right
tdyno.add_structure(shape='wedge',
                    xcenter=xmax - 1., ycenter=5.5, radius=4.5, width=0.5, angle1=90., angle2=270.,
                    kind='dielectric',
                    epsi=12.)

# ============   Source temporal profiles  ============
omega = 2. * np.pi * 0.257
tau = 10.
cw = tdyno.add_source_temporal('cw', omega, tau=tau)

# ============   set up sources   ============
tdyno.add_tfsf_source(xmin=1., xmax=xmax - 2.5, ymin=1., ymax=1.5, kx=1., ky=0., amp=1., t_profile=cw, sides='left', epsi=12.)

# ===============   PML   =================
tdyno.add_pml(poly_ord=3.5, ratio=1.e-7, npx=15, npy=15)

# ============   numeric dispersion correction  ============
tdyno.use_ndc(True, omega)  # the frequency should be that of the source in use

# ==============   Running simulation   =================
tdyno.run(skipping=5, vmin=-1., vmax=1.)



# -*- coding: utf-8 -*-


import numpy as np


class Hm:

    def __init__(self, c, omg, phi=0., t_bf=None, t_r=None):
        """
        Harmonic temporal profile.

        At the beginning, ramp up by a Gaussian profile. After hitting the Gaussian peak, constant amplitude.

        Parameters
        ----------
        c           :   Cnst
                        physical constants
        omg         :   float
                        frequency of the harmonic
        phi         :   float
                        phase shift of the harmonic
        t_bf        :   float
                        total time before the peak of the Gaussian.
        t_r         :   float
                        time-constant of the ramping, i.e. half-width of the Gaussian


        """

        if t_r is None:
            t_r = 10. / c.c0
        # ramp time
        self.t_r = t_r
        # time before
        if t_bf is None:
            self.t_bf = 6. * t_r
        else:
            self.t_bf = t_bf

        self.omg = omg
        self.phi = phi

        # the source field
        self.fld = None

    def f(self, t):

        """
        calculate the field for given times t

        Parameters
        ----------
        t   :   float or ndarray
                for points with t < 0, that's before the beginning of time, return 0
        Returns
        -------

        """
        if type(t) is float:
            t = np.array(t)

        # calculate source field
        self.fld = np.zeros(t.shape)
        i_t_bf = (t > 0.) * (t < self.t_bf)
        self.fld[i_t_bf] = np.exp(- np.square((t[i_t_bf] - self.t_bf) / self.t_r)) * np.sin(self.omg * t[i_t_bf] + self.phi)
        i_t_nm = (t >= self.t_bf)
        self.fld[i_t_nm] = np.sin(self.omg * t[i_t_nm] + self.phi)

        return self.fld


class HP:

    def __init__(self, c, omg, phi=0., t_spn=None, t_bf=None, t_r=None):
        """
        Harmonic Pulse temporal profile.

        At the beginning/end, use a Gaussian to ramp up/down. Constant amplitude in between for t_spn time.

        If t_spn is zero, then this is a Gaussian profile modulated harmonic wave centered at a given frequency omg.

        Parameters
        ----------
        c           :   Cnst
                        physical constants
        omg         :   float
                        frequency of the harmonic wave
        t_spn       :   float
                        length of pulse (the full amplitude section)
        t_bf        :   float
                        total time before the peak of the Gaussian.
        t_r         :   float
                        time constant of the ramping, i.e. half width of the Gaussian


        """

        if t_r is None:
            t_r = 10. / c.c0
        self.t_r = t_r
        if t_bf is None:
            self.t_bf = 6. * t_r
        else:
            self.t_bf = t_bf
        self.t_spn = t_spn

        self.omg = omg
        self.phi = phi

        # the source field
        self.fld = None

    def f(self, t):

        """
        calculate the field for given times t

        Parameters
        ----------
        t   :   float or ndarray
                for points with t < 0, that's before the beginning of time, return 0

        Returns
        -------

        """
        if type(t) is float:
            t = np.array(t)

        # calculate source field
        self.fld = np.zeros(t.shape)
        i_t_bf = (t > 0.) * (t < self.t_bf)
        self.fld[i_t_bf] = np.exp(- np.square((t[i_t_bf] - self.t_bf) / self.t_r)) * np.sin(self.omg * t[i_t_bf] + self.phi)

        i_t_nm = (t >= self.t_bf) * (t <= (self.t_bf + self.t_spn))
        self.fld[i_t_nm] = np.sin(self.omg * t[i_t_nm] + self.phi)
        i_t_af = (t > (self.t_bf + self.t_spn))
        self.fld[i_t_af] = np.exp(- np.square((t[i_t_af] - self.t_bf - self.t_spn) / self.t_r)) * np.sin(self.omg * t[i_t_af] + self.phi)

        return self.fld


class Gsn:

    def __init__(self, c, tau, t_bf=None):
        """
        A pure Gaussian pulse (no carrier frequency).

        Parameters
        ----------
        c       :   Cnst
                    Physical constants
        tau     :   float
                    width of the Gaussian
        t_bf    :   float
                    time before the peak of the Gaussian pulse

        Returns
        -------

        """

        self.tau = tau
        if t_bf is None:
            t_bf = 6. * tau
        self.t_bf = t_bf

        # the source field
        self.fld = None

    def f(self, t):

        """
        calculate the field for given times

        Parameters
        ----------
        t   :   float or ndarray
                for points with t < 0, that's before the beginning of time, return 0

        Returns
        -------

        """

        if type(t) is float:
            t = np.array(t)
        self.fld = np.zeros(t.shape)
        i_t = (t > 0.)
        self.fld[i_t] = np.exp(-np.square((t[i_t] - self.t_bf) / self.tau))

        return self.fld



# -*- coding: utf-8 -*-

import numpy as np
from scipy import optimize as opt
from scipy.interpolate import interp1d
# import time
from warnings import warn
from tdyno.s2t import S2T


class TSSMD2:

    def __init__(self,
                 xmin_ts, xmax_ts, ymin_ts, ymax_ts,
                 kx, ky, amp, g,
                 xi, f, epsi, mu, omg,
                 st, c0, dt,
                 xi0=None, reverse_direction=False,
                 plrz='Ez', if_ndc=False, omg_ndc=None,
                 whr='all',
                 use_epsi_mu_in_structure=False):
        """
        Generate TF/SF plane wave source with given wave vector with supplied temporal profile in the square defined by {xmin, xmax, ymin, ymax}_ts.

        Parameters
        ----------
        xmin_ts, xmax_ts, ymin_ts, ymax_ts :    float
                                                four corners of TFSF source box
        kx, ky          :   float
                            (kx, ky) determine the wave propagation constant. Both the magnitude and the direction of (kx, ky) matters. This is different from `add_tfsf_source`.
        amp             :   float
                            amplitude of the incident wave
        st              :   S2T
        g               :   Hm or Gsn or HP
                            source temporal profile
        xi              :   array_like
                            1d array, spatial points where f is defined, coordinate transverse to the waveguide. See notes on `f` below.
        f               :   array_like
                            1d array value of f, the waveguide mode.

                            For Ez mode, `f` is Ez. For Hz mode, `f` is Hz.

                            `xi` and `f` can be analytically calculated or numerically simulated.

                            `xi` are the actual physical points where `f` is defined.
                            For Ez mode, `xi` is where Ez is defined. For Hz mode, `xi` is where Hz is defined.
                            In fact `xi` does not know about Yee cells in general (for example, analytically simulated).

                            If `xi` and `f` were simulation results of `tdyno` with a waveguide in the x direction, then for `Ez` polarization `xi` are the Yee cell corners i.e. grid points in y, while for `Hz` polarization `xi` are the half grid points in y in each Yee cell.

        epsi            :   array_like
                            2d array. The permittivity profile along the transverse direction of the waveguide.

                            `epsi[i]` defines the values at index i. It has 2 elements, `epsi[i][0]` is the component  along the waveguide direction, `epsi[i][1]` is transverse to the waveguide.

                            See notes in `mu` below.

        mu              :   array_like
                            2d array. The permeability profile along the transverse direction of the waveguide.

                            `mu[i]` defines the values at index i. It has 2 elements, `mu[i][0]` is the component along the waveguide direction, `mu[i][1]` is transverse to the waveguide.

                            For Ez (TM) mode, epsi[i] and mu[i] defines the value in the interval xi[i] to xi[i+1].

                            For Hz (TE) mode, epsi[i] and mu[i] defines the value in the interval (xi[i-1] + xi[i])/2 to (xi[i] + xi[i+1])/2.

                            Such definition is in accordance with Yee-cell definition, i.e. Ez is defined at the corner, Hz is defined at the center.
                            Such definition is in accordance with `tdyno` grid system, i.e. with this `epsi` and `mu`, `tdyno` will calculate the mode `f` at the physical `xi`.

        use_epsi_mu_in_structure : bool
                            If True, use the the permittivity and permeability of the structure being solved.

                            Typically used when `f` is generated with a different resolution.

                            todo: implement this

        omg             :   float
                            intended frequency
        xi0             :   float
                            signed distance from reference line (xi=0) to origin
        reverse_direction : bool
                            If `True`, reverse `beta` direction, while keeping profile `f` unchanged.

                            Can instead manually set kx and ky to negative, in which case it is rotating both `beta` and `f` 180 degrees. Note, for asymmetric waveguide, modal profile `f` is asymmetric.

        plrz            :   str
                            polarization
        if_ndc          :   bool
                            if numeric dispersion compensation

                            If True, do not use ndc in the main update equations.

        omg_ndc         :   float
                            frequency for ndc
        whr             :   str
                            {'all', combinations of 't', 'b', 'l', 'r'}

                            Controls which sides of the TFSF source exist.

                            Default to be 'all', i.e. TFSF source in a rectangular region.

                            Can choose any combination of 't', 'b', 'l', 'r' for top, bottom, left and right.

                            If choose any ONE of the sides, it becomes a uni-directional plane-wave source.
        """

        y = np.array(xi)
        f = np.array(f)
        epsi = np.array(epsi)
        mu = np.array(mu)

        self._xi = y
        # self._xi_h = y_h
        self.f = f
        self.epsi = epsi
        self.mu = mu

        self.reverse_direction = reverse_direction

        dy = y[1] - y[0]
        xp = y[-1] + dy
        xm = y[0] - dy
        yp = np.append(y[1:], xp)
        ym = np.insert(y[:-1], 0, xm)
        yhp = (y + yp) / 2.
        yhm = (y + ym) / 2.
        if plrz == 'Ez' or plrz == 'TM':
            y_h = np.array([y, yhp]).T.ravel()
        else:
            if (plrz != 'Hz') and (plrz != 'TE'):
                warn("polarization not recognized. assume 'Hz'", UserWarning)
            y_h = np.array([yhm, y]).T.ravel()

        self.ff = interp1d(y, f, fill_value="extrapolate")

        epsi_x = epsi[:, 0]
        epsi_y = epsi[:, 1]
        mu_x = mu[:, 0]
        mu_y = mu[:, 1]

        epsi_h = np.array([epsi_y, epsi_y]).T.ravel()
        self.epsif = interp1d(y_h, epsi_h, fill_value="extrapolate")
        mu_h = np.array([mu_y, mu_y]).T.ravel()
        self.muf = interp1d(y_h, mu_h, fill_value="extrapolate")

        fp, fm = self.ff([xp, xm])
        pyff = (np.append(f[1:], fp) - f) / (yp - y)
        pybf = (f - np.insert(f[:-1], 0, fm)) / (y - ym)
        pyffmu = pyff / mu_x
        self.pfmf = interp1d(yhp, pyffmu, fill_value="extrapolate")
        pybfep = pybf / epsi_x
        self.pfef = interp1d(yhm, pybfep, fill_value="extrapolate")

        self.omg = omg
        if xi0 is None:
            self.xi0 = 0.
        else:
            self.xi0 = xi0

        self._c0 = c0
        self._dt = dt
        self._amp = amp
        self._st = st
        self._g = g
        self._plrz = plrz
        self._whr = whr

        self.kx = kx
        self.ky = ky
        self.k = np.sqrt(kx**2 + ky**2)
        k_unit = 1. / np.sqrt(kx**2 + ky**2) * np.array([kx, ky])
        self.kux = k_unit[0]
        self.kuy = k_unit[1]

        self.v = self.omg / self.k * self._c0

        xx_n, yy_n = [st.xx_n, st.yy_n]
        self.xx_n, self.yy_n = (xx_n, yy_n)
        xmin_ts_n, xmax_ts_n, ymin_ts_n, ymax_ts_n = [int(np.floor(cor / dcor)) for cor, dcor in zip([xmin_ts + st.dx / 1e4 - st.xmin, xmax_ts + st.dx / 1e4 - st.xmin, ymin_ts + st.dx / 1e4 - st.ymin, ymax_ts + st.dx / 1e4 - st.ymin], [st.dx, st.dx, st.dy, st.dy])]
        if xmin_ts_n < xx_n.min():
            warn('Source left boundary outside of the solving space. It has been reset to the left edge of the solving space.')
            xmin_ts_n = xx_n.min()
        if xmax_ts_n > xx_n.max():
            warn('Source right boundary outside of the solving space. It has been reset to the right edge of the solving space.')
            xmax_ts_n = xx_n.max()
        if ymin_ts_n < yy_n.min():
            warn('Source bottom boundary outside of the solving space. It has been reset to the bottom edge of the solving space.')
            ymin_ts_n = yy_n.min()
        if ymax_ts_n > yy_n.max():
            warn('Source top boundary outside of the solving space. It has been reset to the top edge of the solving space.')
            ymax_ts_n = yy_n.max()

        self.xmin_ts_n, self.xmax_ts_n, self.ymin_ts_n, self.ymax_ts_n = [xmin_ts_n, xmax_ts_n, ymin_ts_n, ymax_ts_n]

        # indexing from the bulk
        self.i_te_TM = (yy_n == ymax_ts_n) * (xx_n <= xmax_ts_n) * (xx_n > xmin_ts_n)
        self.i_be_TM = (yy_n == ymin_ts_n) * (xx_n <= xmax_ts_n) * (xx_n > xmin_ts_n)
        self.i_le_TM = (xx_n == xmin_ts_n) * (yy_n <= ymax_ts_n) * (yy_n > ymin_ts_n)
        self.i_re_TM = (xx_n == xmax_ts_n) * (yy_n <= ymax_ts_n) * (yy_n > ymin_ts_n)
        # todo: for modal source should always include min exclue max (that's how modal profile should be recorded)
        #
        self.i_te_TE = (yy_n == ymax_ts_n) * (xx_n < xmax_ts_n) * (xx_n >= xmin_ts_n)
        self.i_be_TE = (yy_n == ymin_ts_n) * (xx_n < xmax_ts_n) * (xx_n >= xmin_ts_n)
        self.i_le_TE = (xx_n == xmin_ts_n) * (yy_n < ymax_ts_n) * (yy_n >= ymin_ts_n)
        self.i_re_TE = (xx_n == xmax_ts_n) * (yy_n < ymax_ts_n) * (yy_n >= ymin_ts_n)
        #
        self.i_e_TM = self.i_re_TM + self.i_le_TM + self.i_te_TM + self.i_be_TM
        self.i_e_TE = self.i_re_TE + self.i_le_TE + self.i_te_TE + self.i_be_TE
        #
        self.i_co_tr_TM = self.i_te_TM * self.i_re_TM
        self.i_co_bl_TE = self.i_le_TE * self.i_be_TE

        # self.i_TM_rlv_Dz = self.i_e_TM
        # self.i_TM_rlv_Bx = self.i_te_TM + self.i_be_TM
        # self.i_TM_rlv_By = self.i_le_TM + self.i_re_TM
        # self.i_TE_rlv_Bz = self.i_e_TE
        # self.i_TE_rlv_Dx = self.i_te_TE + self.i_be_TE
        # self.i_TE_rlv_Dy = self.i_le_TE + self.i_re_TE

        self.set_ndc(if_ndc, omg_ndc)

        self._calc_xi()

        self._calc_td()

        # edge indexing
        self.tblr = [False] * 4
        self.iB_s = np.zeros([st.Ny, st.Nx], dtype=bool)
        if whr == "all":
            whr = "tblr"
        if self._plrz == 'Hz':
            for w in whr:
                for i, _w, idx in zip(range(4), "tblr", [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]):
                    if w == _w:
                        self.tblr[i] = True
                        self.iB_s += idx
        elif self._plrz == 'Ez':
            for w in whr:
                for i, _w, idx in zip(range(4), "tblr", [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]):
                    if w == _w:
                        self.tblr[i] = True
                        self.iB_s += idx
        else:
            warn('Polarization not recognized. No TFSF source was set.')
            self.i = [np.zeros([st.Ny, st.Nx], dtype=bool)]
        self.iD_s = self.iB_s

        # prepare source fields
        self.Dx_s, self.Dy_s, self.Dz_s = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]
        self.Bx_s, self.By_s, self.Bz_s = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]

        # TSFS patch
        self.pchs = []
        if whr == 'all':
            xmi = xmin_ts
            ymi = ymin_ts
            width = xmax_ts - xmin_ts
            height = ymax_ts - ymin_ts
            pch = {'shp': 'rct',
                   'xy': (xmi, ymi),
                   'width': width,
                   'height': height
                   }
            self.pchs.append(pch)
        else:
            for w in whr:
                if w == 'l':
                    xmi = xmin_ts
                    ymi = ymin_ts
                    width = 0.
                    height = ymax_ts - ymin_ts
                elif w == 'b':
                    xmi = xmin_ts
                    ymi = ymin_ts
                    width = xmax_ts - xmin_ts
                    height = 0.
                elif w == 'r':
                    xmi = xmax_ts
                    ymi = ymin_ts
                    width = 0.
                    height = ymax_ts - ymin_ts
                elif w == 't':
                    xmi = xmin_ts
                    ymi = ymax_ts
                    width = xmax_ts - xmin_ts
                    height = 0.
                else:
                    warn('TF/SF source location unrecognized.', UserWarning)
                    xmi = ymi = width = height = None

                pch = {'shp': 'rct',
                       'xy': (xmi, ymi),
                       'width': width,
                       'height': height
                       }

                self.pchs.append(pch)

    def set_ndc(self, if_ndc, omg_ndc):
        """
        set numerical dispersion correction

        Parameters
        ----------
        if_ndc
        omg_ndc

        Returns
        -------

        """

        # numeric dispersion
        if if_ndc:
            if omg_ndc:
                omg = omg_ndc
            elif hasattr(self._g, 'omg'):
                omg = self._g.omg
            else:
                omg = self._c0 / .2
            # n = np.sqrt(self.epsi * self.mu)
            n = self.k / self.omg  # effective refractive index

            def func(x):
                return np.square(np.sin(x * self.kux * self._st.dx / 2.) / self._st.dx) + np.square(np.sin(x * self.kuy * self._st.dy / 2.) / self._st.dy) - np.square(np.sin(omg * self._dt / 2.) / self._c0 * n / self._dt)
            k_numeric = opt.brentq(func, (omg / self._c0 * n * 0.5), (omg / self._c0 * n * 2.))
            ndc = omg / self._c0 * n / k_numeric
            # print(ndc)
        else:
            ndc = 1.
        self.ndc = ndc

        self._calc_td()

    def _calc_td(self):
        """
        calculate relative time delays

        Returns
        -------

        """
        ndc = self.ndc

        kux, kuy, kx, ky = [a if not self.reverse_direction else -a for a in [self.kux, self.kuy, self.kx, self.ky]]

        # calculate relative distances
        xx_r = self._st.dx * (self.xx_n - self.xmin_ts_n)
        yy_r = self._st.dy * (self.yy_n - self.ymin_ts_n)
        d_ts = xx_r * kux + yy_r * kuy
        d_ts -= d_ts[self.i_e_TM].min()

        # calculate time delay
        self.td = d_ts / self.v / ndc
        #
        self.tde_Hx = self._st.dy / 2. * ky / self.omg / self._c0 / ndc
        self.tde_Hy = self._st.dx / 2. * kx / self.omg / self._c0 / ndc
        self.tde_Ex = self._st.dx / 2. * kx / self.omg / self._c0 / ndc
        self.tde_Ey = self._st.dy / 2. * ky / self.omg / self._c0 / ndc
        self.tde_Hz = (self._st.dx / 2. * kx + self._st.dy / 2. * ky) / self.omg / self._c0 / ndc
        #
        self.td_TE_Hz = self.td + self.tde_Hz - self._dt / 2.
        self.td_TE_Ex = self.td + self.tde_Ex
        self.td_TE_Ey = self.td + self.tde_Ey
        self.td_TM_Ez = self.td
        self.td_TM_Hx = self.td + self.tde_Hx - self._dt / 2.
        self.td_TM_Hy = self.td + self.tde_Hy - self._dt / 2.
        #
        self.td_TE_Hz_t, self.td_TE_Hz_b, self.td_TE_Hz_l, self.td_TE_Hz_r = [self.td_TE_Hz[idx] for idx in [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]]
        self.td_TE_Ex_t, self.td_TE_Ex_b = [self.td_TE_Ex[idx] for idx in [self.i_te_TE, self.i_be_TE]]
        self.td_TE_Ey_l, self.td_TE_Ey_r = [self.td_TE_Ey[idx] for idx in [self.i_le_TE, self.i_re_TE]]
        self.td_TM_Ez_t, self.td_TM_Ez_b, self.td_TM_Ez_l, self.td_TM_Ez_r = [self.td_TM_Ez[idx] for idx in [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]]
        self.td_TM_Hx_t, self.td_TM_Hx_b = [self.td_TM_Hx[idx] for idx in [self.i_te_TM, self.i_be_TM]]
        self.td_TM_Hy_l, self.td_TM_Hy_r = [self.td_TM_Hy[idx] for idx in [self.i_le_TM, self.i_re_TM]]

        # self.Dx_ic, self.Dy_ic, self.Dz_ic = [np.zeros([_st.Ny, _st.Nx]) for ii in range(3)]
        # self.Bx_ic, self.By_ic, self.Bz_ic = [np.zeros([_st.Ny, _st.Nx]) for ii in range(3)]

    def _calc_xi(self):
        """
        calculate coordinate xi

        Returns
        -------

        """

        # calculate xi to grid
        xx_r = self._st.xx
        yy_r = self._st.yy
        xi1 = yy_r * self.kux - xx_r * self.kuy
        self.xi = xi1 + self.xi0

        # calculate extra distances
        xi_Hx = self._st.dy / 2. * self.kux
        xi_Hy = - self._st.dx / 2. * self.kuy
        xi_Ex = - self._st.dx / 2. * self.kuy
        xi_Ey = self._st.dy / 2. * self.kux
        xi_Hz = - self._st.dx / 2. * self.kuy + self._st.dy / 2. * self.kux
        #
        self.xi_TE_Hz = self.xi + xi_Hz
        self.xi_TE_Ex = self.xi + xi_Ex
        self.xi_TE_Ey = self.xi + xi_Ey
        self.xi_TM_Ez = self.xi
        self.xi_TM_Hx = self.xi + xi_Hx
        self.xi_TM_Hy = self.xi + xi_Hy
        #
        self.xi_TE_Hz_t, self.xi_TE_Hz_b, self.xi_TE_Hz_l, self.xi_TE_Hz_r = [self.xi_TE_Hz[idx] for idx in [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]]
        self.xi_TE_Ex_t, self.xi_TE_Ex_b = [self.xi_TE_Ex[idx] for idx in [self.i_te_TE, self.i_be_TE]]
        self.xi_TE_Ey_l, self.xi_TE_Ey_r = [self.xi_TE_Ey[idx] for idx in [self.i_le_TE, self.i_re_TE]]
        self.xi_TM_Ez_t, self.xi_TM_Ez_b, self.xi_TM_Ez_l, self.xi_TM_Ez_r = [self.xi_TM_Ez[idx] for idx in [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]]
        self.xi_TM_Hx_t, self.xi_TM_Hx_b = [self.xi_TM_Hx[idx] for idx in [self.i_te_TM, self.i_be_TM]]
        self.xi_TM_Hy_l, self.xi_TM_Hy_r = [self.xi_TM_Hy[idx] for idx in [self.i_le_TM, self.i_re_TM]]

    def us(self, t):

        """
        update source fields

        Parameters
        ----------
        t   :   float
                current time

        Returns
        -------

        """

        kux, kuy, kx, ky = [a if not self.reverse_direction else -a for a in [self.kux, self.kuy, self.kx, self.ky]]

        if self._plrz == 'Hz':

            # todo: should there be pi/2 phase delay in Ex?

            if self.tblr[0]:
                f = self.ff(self.xi_TE_Hz_t)
                td = self.td_TE_Hz_t
                Hz_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TE_Ex_t
                fe = self.ff(self.xi_TE_Ex_t) / self.epsif(self.xi_TE_Ex_t)
                pfe = self.pfef(self.xi_TE_Ex_t)
                Ex_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kux
                                                - self._g.f(t - td) * self.k * fe * kuy)
                #
                self.Bz_s[self.i_te_TE] = - self._dt / self._st.dy * Ex_ic
                self.Dx_s[self.i_te_TE] = self._dt / self._st.dy * Hz_ic
            if self.tblr[1]:
                f = self.ff(self.xi_TE_Hz_b)
                td = self.td_TE_Hz_b
                Hz_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TE_Ex_b
                fe = self.ff(self.xi_TE_Ex_b) / self.epsif(self.xi_TE_Ex_b)
                pfe = self.pfef(self.xi_TE_Ex_b)
                Ex_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kux
                                                - self._g.f(t - td) * self.k * fe * kuy)
                Ex_ic_bl_co = Ex_ic[0]
                #
                self.Bz_s[self.i_be_TE] = self._dt / self._st.dy * Ex_ic
                self.Dx_s[self.i_be_TE] = - self._dt / self._st.dy * Hz_ic
            if self.tblr[2]:
                f = self.ff(self.xi_TE_Hz_l)
                td = self.td_TE_Hz_l
                Hz_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TE_Ey_l
                fe = self.ff(self.xi_TE_Ey_l) / self.epsif(self.xi_TE_Ey_l)
                pfe = self.pfef(self.xi_TE_Ey_l)
                Ey_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kuy
                                                + self._g.f(t - td) * self.k * fe * kux)
                #
                self.Bz_s[self.i_le_TE] = self._dt / self._st.dx * Ey_ic
                self.Dy_s[self.i_le_TE] = self._dt / self._st.dx * Hz_ic
                if self.tblr[1]:
                    self.Bz_s[self.i_co_bl_TE] += self._dt / self._st.dy * Ex_ic_bl_co
            if self.tblr[3]:
                f = self.ff(self.xi_TE_Hz_r)
                td = self.td_TE_Hz_r
                Hz_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TE_Ey_r
                fe = self.ff(self.xi_TE_Ey_r) / self.epsif(self.xi_TE_Ey_r)
                pfe = self.pfef(self.xi_TE_Ey_r)
                Ey_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kuy
                                                + self._g.f(t - td) * self.k * fe * kux)
                #
                self.Bz_s[self.i_re_TE] = - self._dt / self._st.dx * Ey_ic
                self.Dy_s[self.i_re_TE] = - self._dt / self._st.dx * Hz_ic

        elif self._plrz == 'Ez':

            if self.tblr[0]:
                f = self.ff(self.xi_TM_Ez_t)
                td = self.td_TM_Ez_t
                Ez_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TM_Hx_t
                fe = self.ff(self.xi_TM_Hx_t) / self.muf(self.xi_TM_Hx_t)
                pfe = self.pfmf(self.xi_TM_Hx_t)
                Hx_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kux
                                                - self._g.f(t - td) * self.k * fe * kuy)
                Hx_ic_tr_co = Hx_ic[-1]
                #
                self.Dz_s[self.i_te_TM] = self._dt / self._st.dy * Hx_ic
                self.Bx_s[self.i_te_TM] = self._dt / self._st.dy * Ez_ic
            if self.tblr[1]:
                f = self.ff(self.xi_TM_Ez_b)
                td = self.td_TM_Ez_b
                Ez_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TM_Hx_b
                fe = self.ff(self.xi_TM_Hx_b) / self.muf(self.xi_TM_Hx_b)
                pfe = self.pfmf(self.xi_TM_Hx_b)
                Hx_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kux
                                                - self._g.f(t - td) * self.k * fe * kuy)
                #
                self.Dz_s[self.i_be_TM] = - self._dt / self._st.dy * Hx_ic
                self.Bx_s[self.i_be_TM] = - self._dt / self._st.dy * Ez_ic
            if self.tblr[2]:
                f = self.ff(self.xi_TM_Ez_l)
                td = self.td_TM_Ez_l
                Ez_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TM_Hy_l
                fe = self.ff(self.xi_TM_Hy_l) / self.muf(self.xi_TM_Hy_l)
                pfe = self.pfmf(self.xi_TM_Hy_l)
                Hy_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kuy
                                                + self._g.f(t - td) * self.k * fe * kux)
                #
                self.Dz_s[self.i_le_TM] = - self._dt / self._st.dx * Hy_ic
                self.By_s[self.i_le_TM] = self._dt / self._st.dx * Ez_ic
            if self.tblr[3]:
                f = self.ff(self.xi_TM_Ez_r)
                td = self.td_TM_Ez_r
                Ez_ic = self._amp * f * self._g.f(t - td)
                #
                td = self.td_TM_Hy_r
                fe = self.ff(self.xi_TM_Hy_r) / self.muf(self.xi_TM_Hy_r)
                pfe = self.pfmf(self.xi_TM_Hy_r)
                Hy_ic = self._amp / self.omg * (self._g.f(t - td - np.pi / 2 / self.omg) * pfe * kuy
                                                + self._g.f(t - td) * self.k * fe * kux)
                #
                self.Dz_s[self.i_re_TM] = self._dt / self._st.dx * Hy_ic
                self.By_s[self.i_re_TM] = - self._dt / self._st.dx * Ez_ic
                if self.tblr[0]:
                    self.Dz_s[self.i_co_tr_TM] += self._dt / self._st.dy * Hx_ic_tr_co





#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
## This is an example FDTD simulation. It contains structures in different shapes and materials. Wave is incident by point sources and a Total-field scattered-field (TF/SF) source. CPML boundaries to absorb the scattered waves.

## What you do: run this file. A figure window is shown, which contains a simple graphical user-interface (GUI):

* Click "play/pause" to start simulation. Click again to pause it.
* Click "step" to run simulation by one time step.
* Click "Render" to show the field in a smoothly rendered plot.
* Click "Reset" to clear everything and start over.
* The Axes plots a rectangular solving space.
  * The gray region at the edge is the CPML.
  * The large purple rectangle marks the TF/SF source boundary.
  * There are several structures in the solving space. Different colors mark different materials.
* Near the top of the figure window are controls for video recording
  * Click "Record" to start recording, click it again to stop recording and write to file.
  * Click "Pause" to pause recording. Then you can still run simulation without recording it.
  * FPS and DPI can be changed
  * You can input your choice of file name for the video.
* Simulation time-steps are displayed.

## The structure of a simulation script:
1. Initialize the solver, choose parameters such as polarization, time step length, solving mode, and total time steps.
2. Add structures to the solving space
3. Setup sources
4. Setup PML if needed
5. Setup monitors if needed
6. Run the file. The GUI is shown.
A detailed example of each part is found in this script.

## Units:
All simulations run in natural units, i.e. permittivity, permeability and light speed are 1.

"""

import numpy as np
from tdyno import TDyno

# ============   set up simulation   ==============
# initiate a simulation
tdyno = TDyno()

# geometry of the solving space
dx = 0.05
xmin = 0.
xmax = 10.
dy = 0.05
ymin = 0.
ymax = 10.

# background permittivity and permeability. They fill the space.
epsi_bg = 1.
mu_bg = 1.

# choose polarization
polarization = 'Ez'  # either 'Hz' or 'Ez'

# time step, if simulation diverges, change into a smaller number
dt = 1. / 2.1 * np.sqrt(dx ** 2 + dy ** 2)

# Total time steps for each simulation.
# After it stops, you can click "play/pause" to continue for another Nt.
Nt = 3000

# solving mode
mode = 'HDE'  # for non-magnetic (permeability is 1 throughout), use 'HDE', otherwise, use 'BHDE'

# Initialize the solver
tdyno.setup(xmin, xmax, dx, ymin, ymax, dy, epsi_bg, mu_bg, polarization=polarization, dt=dt, Nt=Nt, mode=mode)

# ============  add structures   ============
# lossless dispersionless dielectric in a rectangle (black)
xmin_d = 3.
xmax_d = 4.5
ymin_d = 3.
ymax_d = 4.5
epsi_d = 12.
mu_d = 1.
tdyno.add_structure(shape='rectangle',
                    xmin=xmin_d, xmax=xmax_d, ymin=ymin_d, ymax=ymax_d,
                    kind='dielectric',
                    epsi=epsi_d, mu=mu_d)

# lossy dispersionless dielectric in a ring (green)
# need the following material parameters
# `conductivity`, positive means loss, negative means gain
xcenter = 3.7
ycenter = 6.4
radius = 0.8
width = 0.3
epsi_b = 2.25
mu_b = 1.
conductivity = 0.5
tdyno.add_structure(shape='ring',
                    xcenter=xcenter, ycenter=ycenter, radius=radius, width=width,
                    kind='dielectric',
                    epsi=epsi_b, mu=mu_b, conductivity=conductivity)

# Lorentz dielectric in a disk  (blue)
# need the following material parameters
# `epsi`, `mu`, permittivity and permeability at high frequencies
# `res_omegas`, the resonance frequencies
# `d_epsi`, the change in permittivity at each resonance, i.e. the resonance strength
# `gammas`, broadening for each resonances
xcenter = 6.4
ycenter = 3.7
radius = 0.8
epsi_Lz = 1.5
mu_Lz = 1.
res_omegas = [2 * np.pi / wl for wl in [8., 4., 12.]]
d_epsi = [12., 11., 2.]
gammas = [2 * np.pi / wl for wl in [12., 10., 3.]]
tdyno.add_structure(shape='disk',
                    xcenter=xcenter, ycenter=ycenter, radius=radius,
                    kind='Lorentz',
                    epsi=epsi_Lz, mu=mu_Lz, res_omegas=res_omegas, d_epsi=d_epsi, gammas=gammas)

# Drude metal in a wedge (yellow)
# need the following parameters
# `epsi`, `mu`, permittivity and permeability at high frequencies
# `res_omegas`, the resonance frequencies
# `gammas`, broadening for the resonances
xcenter = 6.4
ycenter = 6.4
radius = 0.8
width = 0.3
angle1 = 0.
angle2 = 270.
epsi_b = 1.
mu_b = 1.
res_omegas = [2 * np.pi / wl for wl in [1.]]
gammas = [2 * np.pi / wl for wl in [2.]]
tdyno.add_structure(shape='wedge',
                    xcenter=xcenter, ycenter=ycenter, radius=radius, width=width, angle1=angle1, angle2=angle2,
                    kind='metal',
                    epsi=epsi_b, mu=mu_b, res_omegas=res_omegas, gammas=gammas)

# ============   Source temporal profiles  ============
# The temporal profile can be any of the following
# `cw` (continuous wave),
# `pulse` (single Gaussian pulse),
# `packet` (cw wave modulated by a Gaussian profile)

# Gaussian pulse
tau = 2.  # time constant of the pulse
pulse = tdyno.add_source_temporal('pulse', tau=tau)

# cw with single frequency
# initially the wave must ramps up by a Gaussian profile. After reaching the peak, the amplitude stays constant.
wl = 2.  # wavelength
omega = 2. * np.pi / wl  # frequency
phi = np.pi / 2.  # phase
tau = 4.  # how fast the wave ramps up.
t_before = 45.  # total time before reaching steady amplitude
cw = tdyno.add_source_temporal('cw', omega=omega, phase=phi, tau=tau, t_before=t_before)

# a wave packet, harmonic wave (carrier) modulated by a Gaussian profile
wl = 2.  # carrier wavelength
omega = 2. * np.pi / wl  # carrier frequency
t_span = 0.  # after reaching the Gaussian peak, keep amplitude at this level for t_spn amount of time
tau = 4.  # how fast the wave ramps up.
t_before = 25.  # total time before reaching peak amplitude
packet = tdyno.add_source_temporal('packet', omega=omega, t_span=t_span, tau=tau, t_before=t_before)

# ============   set up sources   ============
# After setting up the temporal profile, now set up the source
# A Point source
x_sc = 7.5  # positions
y_sc = 7.5
amp = 100.  # amplitude of the source
tdyno.add_point_source(x=x_sc, y=y_sc, amp=amp, t_profile=pulse)

# A Point source
x_sc = 2.5  # positions
y_sc = 7.5
amp = 20  # amplitude of the source
tdyno.add_point_source(x=x_sc, y=y_sc, amp=amp, t_profile=packet)

# TFSF Source
xmin_ts = 2.  # the corners of the TFSF source
xmax_ts = 8
ymin_ts = 2.
ymax_ts = 8
# kx and ky only determine the wave direction. The magnitude of the wave vector in vacuum is automatically set by the frequency in the temporal profile.
kx = 0.25
ky = 0.5
amp = 0.5  # wave amplitude
# `sides` defaults to be `'all'`. Can also be 'top', 'bottom', 'left' or 'right' for uni-directional plane wave sources.
sides = 'all'
tdyno.add_tfsf_source(xmin=xmin_ts, xmax=xmax_ts, ymin=ymin_ts, ymax=ymax_ts, kx=kx, ky=ky, amp=amp, t_profile=cw, sides=sides)

# ============   PML   ============
poly_order = 3.5  # pml polynomial order
ratio = 1.e-7  # target attenuation ratio
npx = 15  # number of PML cells in x and y direction
npy = 15
tdyno.add_pml(poly_order, ratio, npx=npx, npy=npy)

# ============   numeric dispersion correction  ============
# choose if you need numeric dispersion correction (ndc).
# If True, need to supply the frequency for ndc.
tdyno.use_ndc(True, omega)  # the frequency should be that of the source in use

# ============   Running simulation   ============
skipping = 5  # skip some frames in plotting (save your time)
tdyno.run(skipping=skipping, vmin=-1., vmax=1.)  # vmin and vmax set the data range for plotting




# -*- coding: utf-8 -*-


class Cnst:

    def __init__(self, c0=1., mu0=1., epsi0=1.):
        """
        Physical constants.

        These constants are default to 1.

        c0 = 1 / np.sqrt(epsi0 * mu0) has to hold, otherwise PML will not function!

        Parameters
        ----------
        c0
        mu0
        epsi0
        """

        self.c0 = c0
        self.mu0 = mu0
        self.epsi0 = epsi0



# -*- coding: utf-8 -*-


import numpy as np
import numpy.linalg as la
from warnings import warn
from typing import List, Union


class S2T:

    def __init__(self, xmin, xmax, dx, ymin, ymax, dy, epsi_bg, mu_bg, x=None, y=None):

        """
        Set up the structure to be solved in 2D FDTD.

        Parameters
        ----------
        xmin, xmax, dx, ymin, ymax, dy      :   float
        epsi_bg, mu_bg                      :   Union[float, np.ndarray]
        x, y                                :   ndarray
                                                1d ndarray, If supplied, override xmin etc.
        """

        if type(epsi_bg) is float:
            self.epsi_bg = np.eye(3, 3) * epsi_bg
        elif np.array(epsi_bg).ndim == 1:
            self.epsi_bg = np.diag(epsi_bg)
        else:
            self.epsi_bg = epsi_bg
        if type(mu_bg) is float:
            self.mu_bg = np.eye(3, 3) * mu_bg
        elif np.array(mu_bg).ndim == 1:
            self.mu_bg = np.diag(mu_bg)
        else:
            self.mu_bg = mu_bg
        self.epsi_bg_inv = la.inv(self.epsi_bg)
        self.mu_bg_inv = la.inv(self.mu_bg)

        # parameters of the solving space
        if x is not None:
            dx = x[1] - x[0]
            xmin = x[0]
            xmax = x[-1] + dx
            Nx = x.size
            x_n = np.arange(Nx, dtype=int)
        else:
            Nx = int((xmax + dx / 1.e4 - xmin) / dx)
            x_n = np.arange(Nx, dtype=int)
            x = x_n * dx + xmin
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.xres = dx
        self.Nx = Nx
        self.x_n = x_n
        self.x = x
        #
        if y is not None:
            dy = y[1] - y[0]
            ymin = y[0]
            ymax = y[-1] + dy
            Ny = y.size
            y_n = np.arange(Nx, dtype=int)
        else:
            Ny = int((ymax + dx / 1.e4 - ymin) / dy)
            y_n = np.arange(Ny, dtype=int)
            y = y_n * dy + ymin
        self.ymin = ymin
        self.ymax = ymax
        self.dy = dy
        self.yres = dy
        self.Ny = Ny
        self.y_n = y_n
        self.y = y
        #
        self.xx_n, self.yy_n = np.meshgrid(self.x_n, self.y_n)
        self.xx, self.yy = np.meshgrid(self.x, self.y)

        # list of box objects
        self.bxs: List[Bx2] = []
        self.bxs_sp: List[Bx2] = []
        self.bxs_Lz: List[Bx2] = []
        self.bxs_dmLz: List[Bx2] = []
        self.bxs_Dr: List[Bx2] = []
        self.bxs_dmri: List[Bx2] = []
        self.bxs_lg: List[Bx2] = []
        self.bxs_dmlg: List[Bx2] = []

    def a_b(self,
            shp='rct',
            xmin_b=None, xmax_b=None, ymin_b=None, ymax_b=None,
            xc_b=None, yc_b=None,
            r_b=None,
            a_1=None, a_2=None,
            w_b=None,
            knd='sp',
            epsi_b=None, mu_b=None,
            epsi_infty_b=None, mu_infty_b=None,
            omgs_rsn=None, dlts_epsi=None, Gms=None,
            sgm=None,
            m_a=None, m_o=None, m_q=None, m_p=0.):

        """
        Add boxes to the structure.

        Parameters
        ----------
        shp                             :   str
                                            ['rct'|'ccl'|'rng'|'wdg']
        xmin_b, xmax_b, ymin_b, ymax_b  :   float
                                            four corners of the box.
        xc_b, yc_b                      :   float
                                            x and y center of box
        r_b                             :   float
                                            radius of circular box, outer radius for ring and wedge box
        a_1, a_2                        :   float
                                            start and end angle of wedge, in degrees, between 0 to 360
        w_b                             :   float
                                            width of ring or wedge box
        knd                             :   str
                                            ['sp'|'Lz'|'dmLz'|'Dr'|'lg'|'dmri'|'dmlg']
        epsi_b, mu_b                    :   float or Tuple[float, float, float] or 3x3 ndarray
                                            The permittivity and permeability.
        epsi_infty_b, mu_infty_b        :   float or Tuple[float, float, float] or 3x3 ndarray
                                            The high-frequency permittivity and permeability.
                                            If not supplied, fall back to epsi_b and mu_b.
        omgs_rsn                        :   list[float]
                                            resonance frequencies
        dlts_epsi                       :   list[float]
                                            permittivity change at each resonance
        Gms                             :   list[float]
                                            Broadening of each resonance
        sgm                             :   float
                                            loss (positive number) or gain (negative) in the material
        m_a                             :   float or list[float]
                                            modulation amplitude
        m_o                             :   float or list[float]
                                            modulation frequency
        m_q                             :   Tuple[float, float] or list[Tuple[float, float]]
                                            (qx, qy), wave vector of modulation.
        m_p                             :   float or list[float]
                                            modulation phase
                                            By default, cos modulation.

        Returns
        -------

        """

        gmt = Gmt2(self.xx, self.yy,
                   shp=shp,
                   xmin_rec_b=xmin_b, xmax_rec_b=xmax_b, ymin_rec_b=ymin_b, ymax_rec_b=ymax_b,
                   xc_b=xc_b, yc_b=yc_b,
                   r_b=r_b,
                   a_1=a_1, a_2=a_2,
                   w_b=w_b)

        mtr = Mtr(knd=knd,
                  epsi=epsi_b, mu=mu_b, epsi_infty=epsi_infty_b, mu_infty=mu_infty_b,
                  omgs_rsn=omgs_rsn, dlts_epsi=dlts_epsi, Gms=Gms,
                  sgm=sgm,
                  m_a=m_a, m_o=m_o, m_q=m_q, m_p=m_p)

        bx = Bx2(gmt, mtr)

        if knd == 'sp':
            self.bxs_sp.append(bx)
        elif knd == 'Lz':
            self.bxs_Lz.append(bx)
        elif knd == 'dmLz':
            self.bxs_dmLz.append(bx)
        elif knd == 'Dr':
            self.bxs_Dr.append(bx)
        elif knd == 'dmri':
            self.bxs_dmri.append(bx)
        elif knd == 'lg':
            self.bxs_lg.append(bx)
        elif knd == 'dmlg':
            self.bxs_dmlg.append(bx)
        else:
            print('Material not recognized.')

        if bx is not None:
            self.bxs.append(bx)


class Bx2:
    def __init__(self, gmt, mtr):
        """
        A box in 2D.

        Parameters
        ----------
        gmt        :   Gmt2
        mtr        :   Mtr
        """

        self.gmt = gmt
        self.mtr = mtr


class MtrBsc(object):
    def __init__(self, epsi=None, mu=None, **kwargs):
        """
        Basic material.

        Parameters
        ----------
        epsi, mu    :   float | (float) | numpy.ndarray
                        permittivity and permeability. if tuple, it is length of 3. if ndarray, it has shape (3, 3)
        kwargs      :
                        other kwargs not related to materials
        """
        if epsi is None:
            epsi = 1.
        if mu is None:
            mu = 1.

        if not hasattr(epsi, '__len__'):
            epsi = np.eye(3, 3) * epsi
        elif np.array(epsi).ndim == 1:
            epsi = np.diag(epsi)
        if not hasattr(mu, '__len__'):
            mu = np.eye(3, 3) * mu
        elif np.array(mu).ndim == 1:
            mu = np.diag(mu)
        self.epsi = epsi
        self.mu = mu


class Mtr(MtrBsc):
    def __init__(self,
                 knd='sp',
                 epsi_infty=1., mu_infty=1., omgs_rsn=None, dlts_epsi=None, Gms=None,
                 sgm=None,
                 m_a=None, m_o=None, m_q=None, m_p=None,
                 **kw):
        """
        Material type and parameters.

        Parameters
        ----------
        knd                             :   str
                                            Options are:
                                            'sp' (simple),
                                            'Lz' (Lorentz),
                                            'dmLz' (dynamic modulation of delta epsi),
                                            'Dr' (Drude),
                                            'lg' (loss and gain),
                                            'dmri' (dynamic modulation of real part of index),
                                            'dmlg' (dynamic modulation of loss and gain).
        epsi_infty_b                    :   float | (float) | numpy.ndarray
        mu_infty_b                      :   float | (float) | numpy.ndarray
                                            if tuple it has len 3, if ndarray it has shape (3, 3).

                                            The high-frequency permittivity and permeability for use in Lorentz and Drude.

                                            If not supplied, fall back to `epsi_b` and `mu_b`.

        omgs_rsn                        :   list[float]
                                            resonance frequencies, for use in Lorentz or Drude
        dlts_epsi                       :   list[float]
                                            permittivity change at each resonance, for use in Lorentz
        Gms                             :   list[float]
                                            Broadening of each resonance, for use in Lorentz or Drude
        sgm                             :   float
                                            loss (positive number) or gain (negative) in the material
        m_a                             :   float or list[float]
                                            modulation amplitude
        m_o                             :   float or list[float]
                                            modulation frequency
        m_q                             :   Tuple[float, float] or list[Tuple[float, float]]
                                            (qx, qy), wave vector of modulation.
        m_p                             :   float or list[float]
                                            modulation phase
                                            By default, cos modulation.

        Keyword Arguments
        -----------------
        kw                      :   keyword arguments to pass to MtrBsc
        """

        if (knd == 'Lz') or (knd == 'dmLz') or (knd == 'Dr'):
            if epsi_infty is None:
                epsi_infty = kw.pop('epsi')
            if mu_infty is None:
                mu_infty = kw.pop('mu')
            super(Mtr, self).__init__(epsi=epsi_infty, mu=mu_infty)
            self.epsi_infty = self.epsi
            self.mu_infty = self.mu
        else:
            super(Mtr, self).__init__(**kw)

        self.knd = knd

        if omgs_rsn is None:
            omgs_rsn = []
        if dlts_epsi is None:
            dlts_epsi = []
        if Gms is None:
            Gms = []

        self.omgs_rsn = omgs_rsn
        self.dlts_epsi = dlts_epsi
        self.Gms = Gms

        self.sgm = sgm

        self.m_a = m_a
        self.m_o = m_o
        self.m_q = m_q
        self.m_p = m_p


class Gmt2:
    def __init__(self, xx, yy,
                 shp='rct',
                 xmin_rec_b=None, xmax_rec_b=None, ymin_rec_b=None, ymax_rec_b=None,
                 xc_b=None, yc_b=None,
                 r_b=None,
                 w_b=None,
                 a_1=None, a_2=None,
                 **kwargs
                 ):

        """
        Geometrical information of a box.

        Parameters
        ----------
        xx, yy          :   ndarray
        shp             :   str
                            'rct' (rectangular),
                            'ccl' (circular),
                            'rng' (ring),
                            'wdg' (wedge)
        xmin_rec_b
        xmax_rec_b
        ymin_rec_b
        ymax_rec_b      :   float
                            four corners of a rectangular box.
        xc_b, yc_b      :   float
                            x and y center of box
        r_b             :   float
                            radius of circular box, the outer radius for ring and wedge box
        a_1, a_2        :   float
                            the start and the end angle of an wedge, in degrees, between 0 to 360
        w_b             :   float
                            width of a ring or wedge box
        kwargs          :
                            other kwargs supplied but not related to geometry
        """

        # self.shp = shp
        if shp == 'rct':
            self.pch = {'shp': 'rct',
                        'xy': (xmin_rec_b, ymin_rec_b),
                        'width': xmax_rec_b - xmin_rec_b,
                        'height': ymax_rec_b - ymin_rec_b
                        }
            # indexing in box
            self.iib = (xx < xmax_rec_b) * (xx >= xmin_rec_b) * (yy < ymax_rec_b) * (yy >= ymin_rec_b)
            self.iibV = self.iib.ravel()

        elif shp == 'ccl':
            self.pch = {'shp': 'ccl',
                        'center': (xc_b, yc_b),
                        'radius': r_b,
                        }
            # indexing in box
            d = np.sqrt(((xx - xc_b) ** 2 + (yy - yc_b) ** 2))
            self.iib = d < r_b
            self.iibV = self.iib.ravel()

        elif shp == 'rng':
            self.pch = {'shp': 'rng',
                        'center': (xc_b, yc_b),
                        'radius': r_b,
                        'width': w_b,
                        }
            # indexing in box
            d = np.sqrt(((xx - xc_b) ** 2 + (yy - yc_b) ** 2))
            self.iib = (d < r_b) * (d >= r_b - w_b)
            self.iibV = self.iib.ravel()

        elif shp == 'wdg':
            self.pch = {'shp': 'wdg',
                        'center': (xc_b, yc_b),
                        'radius': r_b,
                        'angles': (a_1, a_2),
                        'width': w_b,
                        }
            # indexing in box
            d = np.sqrt(((xx - xc_b) ** 2 + (yy - yc_b) ** 2))
            if w_b is None:
                i_i_r = d < r_b
            else:
                i_i_r = (d < r_b) * (d >= r_b - w_b)
            z_ = (xx - xc_b) + (yy - yc_b) * 1j
            a_ = np.angle(z_, deg=True)
            a1 = (a_1 + 180) % 360 - 180
            # regularize angles
            if a_2 != 180.:
                a2 = (a_2 + 180) % 360 - 180
            else:
                a2 = 180.
            if a2 > a1:
                i_i_a = (a_ >= a1) * (a_ < a2)
            else:
                i_i_a = np.logical_not(((a_ >= a1) * (a_ < a2)))
            self.iib = i_i_r * i_i_a
            self.iibV = self.iib.ravel()



# -*- coding: utf-8 -*-

# todo: anisotropic

import numpy as np
from scipy import optimize as opt
import time
from warnings import warn


class PS2:

    def __init__(self, x_s, y_s, amp, g, st, plrz='Hz'):

        """
        Generate a point source at a given location with a given temporal profile.

        Parameters
        ----------
        x_s,y_s :   float
                    x and y location of source
        amp     :   float
                    amplitude of the point source
        g       :   Union[Hm, HP, Gsn]
                    Temporal profile
        st      :   S2T
                    structure
        plrz    :   str
                    polarization

        """

        self.amp = amp
        self.g = g
        self.polarization = plrz

        if x_s > st.xmax:
            x_s = st.xmax
        if x_s < st.xmin:
            x_s = st.min
        if y_s > st.ymax:
            y_s = st.ymax
        if y_s < st.ymin:
            y_s = st.ymin
        x_s_n = int((x_s - st.xmin) / st.dx)
        y_s_n = int((y_s - st.ymin) / st.dy)

        self.iB_s = (st.xx_n == x_s_n) * (st.yy_n == y_s_n)
        self.iD_s = (st.xx_n == x_s_n) * (st.yy_n == y_s_n)

        self.Bx_s_a, self.By_s_a, self.Bz_s_a = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]
        self.Dx_s_a, self.Dy_s_a, self.Dz_s_a = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]

        self.Bx_s, self.By_s, self.Bz_s = [f for f in [self.Bx_s_a, self.By_s_a, self.Bz_s_a]]
        self.Dx_s, self.Dy_s, self.Dz_s = [f for f in [self.Dx_s_a, self.Dy_s_a, self.Dz_s_a]]

    def us(self, t):

        """
        update source

        Parameters
        ----------
        t   :   float
                time, not in time step, but in physical time.

        """

        if self.polarization == 'Hz':
            self.Bz_s[self.iB_s] = self.amp * self.g.f(t)
        elif self.polarization == 'Ez':
            self.Dz_s[self.iD_s] = self.amp * self.g.f(t)


class TSS2:

    def __init__(self,
                 xmin_ts, xmax_ts, ymin_ts, ymax_ts,
                 kx, ky, amp, g,
                 st, c0, dt, plrz='Ez', if_ndc=False, omg_ndc=None,
                 epsi=1., mu=1.,
                 whr='all'):

        """
        Generate TF/SF plane wave source with given wave vector with supplied temporal profile in the square defined by {xmin, xmax, ymin, ymax}_ts.

        Parameters
        ----------
        xmin_ts, xmax_ts, ymin_ts, ymax_ts  :   float
                                                four corners of TFSF source box
        kx, ky          :   float
        amp             :   float
                            amplitude of the incident wave
        st              :   S2T
        g               :   Union[Hm, HP, Gsn]
                            source temporal profile
        plrz            :   str
                            polarization
        if_ndc          :   bool
                            if numeric dispersion compensation
        omg_ndc         :   float
                            frequency for ndc
        epsi, mu        :   float
                            background relative permittivity and permeability, if the src is in a material rather than in vacuum
        whr             :   str
                            {'all', combinations of 't', 'b', 'l', 'r'}

                            Controls which sides of the TFSF source exist.

                            Default to be 'all', i.e. TFSF source in a rectangular region.

                            Can choose any combination of 't', 'b', 'l', 'r' for top, bottom, left and right.

                            If choose any ONE of the sides, it becomes a uni-directional plane-wave source.
        """

        self._epsi = epsi
        self._mu = mu
        self._n = np.sqrt(epsi * mu)
        self._c0 = c0
        self._dt = dt
        self._amp = amp
        self._st = st
        self._g = g
        self._plrz = plrz
        self._whr = whr

        if plrz == 'Ez':
            ymin_ts -= st.dy
            ymax_ts -= st.dy
            xmin_ts -= st.dx
            xmax_ts -= st.dx

        self.kux, self.kuy = 1. / np.sqrt(kx**2 + ky**2) * np.array([kx, ky])

        xx_n, yy_n = [st.xx_n, st.yy_n]
        self.xx_n, self.yy_n = (xx_n, yy_n)
        xmin_ts_n, xmax_ts_n, ymin_ts_n, ymax_ts_n = [int(np.floor(cor / dcor)) for cor, dcor in zip([xmin_ts + st.dx / 1e4 - st.xmin, xmax_ts + st.dx / 1e4 - st.xmin, ymin_ts + st.dx / 1e4 - st.ymin, ymax_ts + st.dx / 1e4 - st.ymin], [st.dx, st.dx, st.dy, st.dy])]
        if xmin_ts_n < xx_n.min():
            warn('Source left boundary outside of the solving space. It has been reset to the left edge of the solving space.')
            xmin_ts_n = xx_n.min()
        if xmax_ts_n > xx_n.max():
            warn('Source right boundary outside of the solving space. It has been reset to the right edge of the solving space.')
            xmax_ts_n = xx_n.max()
        if ymin_ts_n < yy_n.min():
            warn('Source bottom boundary outside of the solving space. It has been reset to the bottom edge of the solving space.')
            ymin_ts_n = yy_n.min()
        if ymax_ts_n > yy_n.max():
            warn('Source top boundary outside of the solving space. It has been reset to the top edge of the solving space.')
            ymax_ts_n = yy_n.max()
        self.xmin_ts_n, self.xmax_ts_n, self.ymin_ts_n, self.ymax_ts_n = [xmin_ts_n, xmax_ts_n, ymin_ts_n, ymax_ts_n]

        # indexing of the edge in the bulk
        # all 2d arrays
        self.i_te_TM = (yy_n == ymax_ts_n) * (xx_n <= xmax_ts_n) * (xx_n > xmin_ts_n)
        self.i_be_TM = (yy_n == ymin_ts_n) * (xx_n <= xmax_ts_n) * (xx_n > xmin_ts_n)
        self.i_le_TM = (xx_n == xmin_ts_n) * (yy_n <= ymax_ts_n) * (yy_n > ymin_ts_n)
        self.i_re_TM = (xx_n == xmax_ts_n) * (yy_n <= ymax_ts_n) * (yy_n > ymin_ts_n)
        #
        self.i_te_TE = (yy_n == ymax_ts_n) * (xx_n < xmax_ts_n) * (xx_n >= xmin_ts_n)
        self.i_be_TE = (yy_n == ymin_ts_n) * (xx_n < xmax_ts_n) * (xx_n >= xmin_ts_n)
        self.i_le_TE = (xx_n == xmin_ts_n) * (yy_n < ymax_ts_n) * (yy_n >= ymin_ts_n)
        self.i_re_TE = (xx_n == xmax_ts_n) * (yy_n < ymax_ts_n) * (yy_n >= ymin_ts_n)
        #
        self.i_e_TM = self.i_re_TM + self.i_le_TM + self.i_te_TM + self.i_be_TM
        self.i_e_TE = self.i_re_TE + self.i_le_TE + self.i_te_TE + self.i_be_TE
        #
        self.i_co_tr_TM = self.i_te_TM * self.i_re_TM
        self.i_co_bl_TE = self.i_le_TE * self.i_be_TE

        # edge indexing
        self.tblr = [False] * 4
        self.iB_s = np.zeros([st.Ny, st.Nx], dtype=bool)
        # self.iD_s = np.zeros([_st.Ny, _st.Nx], dtype=bool)
        if whr == "all":
            whr = "tblr"
        if self._plrz == 'Hz':
            for w in whr:
                for i, _w, idx in zip(range(4), "tblr", [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]):
                    if w == _w:
                        self.tblr[i] = True
                        self.iB_s += idx
        elif self._plrz == 'Ez':
            for w in whr:
                for i, _w, idx in zip(range(4), "tblr", [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]):
                    if w == _w:
                        self.tblr[i] = True
                        self.iB_s += idx
        else:
            raise Exception('Polarization not recognized. No TF/SF source was set.')
        self.iD_s = self.iB_s

        #
        self.Dx_s, self.Dy_s, self.Dz_s = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]
        self.Bx_s, self.By_s, self.Bz_s = [np.zeros([st.Ny, st.Nx]) for ii in range(3)]

        # TSFS patch
        self.pchs = []
        if whr == 'all':
            xmi = xmin_ts
            ymi = ymin_ts
            width = xmax_ts - xmin_ts
            height = ymax_ts - ymin_ts
            pch = {'shp': 'rct',
                   'xy': (xmi, ymi),
                   'width': width,
                   'height': height
                   }
            self.pchs.append(pch)
        else:
            for w in whr:
                if w == 'l':
                    xmi = xmin_ts
                    ymi = ymin_ts
                    width = 0.
                    height = ymax_ts - ymin_ts
                elif w == 'b':
                    xmi = xmin_ts
                    ymi = ymin_ts
                    width = xmax_ts - xmin_ts
                    height = 0.
                elif w == 'r':
                    xmi = xmax_ts
                    ymi = ymin_ts
                    width = 0.
                    height = ymax_ts - ymin_ts
                elif w == 't':
                    xmi = xmin_ts
                    ymi = ymax_ts
                    width = xmax_ts - xmin_ts
                    height = 0.
                else:
                    warn('TF/SF source location unrecognized.', UserWarning)
                    xmi = ymi = width = height = None

                pch = {'shp': 'rct',
                       'xy': (xmi, ymi),
                       'width': width,
                       'height': height
                       }

                self.pchs.append(pch)

        self.set_ndc(if_ndc, omg_ndc)

        # self._calc_td()  # called in `self.set_ndc`

    def set_ndc(self, if_ndc, omg_ndc):
        """
        set numerical dispersion correction

        Parameters
        ----------
        if_ndc
        omg_ndc

        Returns
        -------

        """
        # numeric dispersion
        if if_ndc:
            if omg_ndc:
                omg = omg_ndc
            elif hasattr(self._g, 'omg'):
                omg = self._g.omg
            else:
                omg = self._c0 / .2
            n = np.sqrt(self._epsi * self._mu)

            def func(x):
                return np.square(np.sin(x * self.kux * self._st.dx / 2.) / self._st.dx) + np.square(np.sin(x * self.kuy * self._st.dy / 2.) / self._st.dy) - np.square(np.sin(omg * self._dt / 2.) / self._c0 * n / self._dt)
            k_numeric = opt.brentq(func, (omg / self._c0 * n * 0.5), (omg / self._c0 * n * 2.))
            ndc = omg / self._c0 * n / k_numeric
            # print(ndc)
        else:
            ndc = 1.
        self.ndc = ndc

        self._calc_td()

    def _calc_td(self):
        """
        calculate relative time delays

        Returns
        -------

        """
        ndc = self.ndc

        # calculate relative distances
        xx_r = self._st.dx * (self.xx_n - self.xmin_ts_n)
        yy_r = self._st.dy * (self.yy_n - self.ymin_ts_n)
        d_ts = xx_r * self.kux + yy_r * self.kuy
        d_ts -= d_ts[self.i_e_TM].min()

        # calculate time delay
        self.td = d_ts / self._c0 * self._n / ndc
        #
        self.tde_Hx = self._st.dy / 2. * self.kuy / self._c0 * self._n / ndc
        self.tde_Hy = self._st.dx / 2. * self.kux / self._c0 * self._n / ndc
        self.tde_Ex = self._st.dx / 2. * self.kux / self._c0 * self._n / ndc
        self.tde_Ey = self._st.dy / 2. * self.kuy / self._c0 * self._n / ndc
        self.tde_Hz = (self._st.dx / 2. * self.kux + self._st.dy / 2. * self.kuy) / self._c0 * self._n / ndc
        #
        self.td_TE_Hz = self.td + self.tde_Hz - self._dt / 2.
        self.td_TE_Ex = self.td + self.tde_Ex
        self.td_TE_Ey = self.td + self.tde_Ey
        self.td_TM_Ez = self.td
        self.td_TM_Hx = self.td + self.tde_Hx - self._dt / 2.
        self.td_TM_Hy = self.td + self.tde_Hy - self._dt / 2.
        #
        self.td_TE_Hz_t, self.td_TE_Hz_b, self.td_TE_Hz_l, self.td_TE_Hz_r = [self.td_TE_Hz[idx] for idx in [self.i_te_TE, self.i_be_TE, self.i_le_TE, self.i_re_TE]]
        self.td_TE_Ex_t, self.td_TE_Ex_b = [self.td_TE_Ex[idx] for idx in [self.i_te_TE, self.i_be_TE]]
        self.td_TE_Ey_l, self.td_TE_Ey_r = [self.td_TE_Ey[idx] for idx in [self.i_le_TE, self.i_re_TE]]
        self.td_TM_Ez_t, self.td_TM_Ez_b, self.td_TM_Ez_l, self.td_TM_Ez_r = [self.td_TM_Ez[idx] for idx in [self.i_te_TM, self.i_be_TM, self.i_le_TM, self.i_re_TM]]
        self.td_TM_Hx_t, self.td_TM_Hx_b = [self.td_TM_Hx[idx] for idx in [self.i_te_TM, self.i_be_TM]]
        self.td_TM_Hy_l, self.td_TM_Hy_r = [self.td_TM_Hy[idx] for idx in [self.i_le_TM, self.i_re_TM]]

        # self.Dx_ic, self.Dy_ic, self.Dz_ic = [np.zeros([_st.Ny, _st.Nx]) for ii in range(3)]
        # self.Bx_ic, self.By_ic, self.Bz_ic = [np.zeros([_st.Ny, _st.Nx]) for ii in range(3)]

    def us(self, t):

        """
        update source fields

        Parameters
        ----------
        t   :   float
                current time

        Returns
        -------

        """

        if self._plrz == 'Hz':
            # ======  method 3  ======
            if self.tblr[0]:
                Hz_ic = self._amp * self._g.f(t - self.td_TE_Hz_t)
                Ex_ic = self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TE_Ex_t) * self.kuy
                self.Bz_s[self.i_te_TE] = - self._dt / self._st.dy * Ex_ic
                self.Dx_s[self.i_te_TE] = self._dt / self._st.dy * Hz_ic
            if self.tblr[1]:
                Hz_ic = self._amp * self._g.f(t - self.td_TE_Hz_b)
                Ex_ic = self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TE_Ex_b) * self.kuy
                Ex_ic_bl_co = Ex_ic[0]
                self.Bz_s[self.i_be_TE] = self._dt / self._st.dy * Ex_ic
                self.Dx_s[self.i_be_TE] = - self._dt / self._st.dy * Hz_ic
            if self.tblr[2]:
                Hz_ic = self._amp * self._g.f(t - self.td_TE_Hz_l)
                Ey_ic = self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TE_Ey_l) * self.kux
                self.Bz_s[self.i_le_TE] = self._dt / self._st.dx * Ey_ic  # - sign?
                self.Dy_s[self.i_le_TE] = self._dt / self._st.dx * Hz_ic
                if self.tblr[1]:
                    self.Bz_s[self.i_co_bl_TE] += self._dt / self._st.dy * Ex_ic_bl_co
            if self.tblr[3]:
                Hz_ic = self._amp * self._g.f(t - self.td_TE_Hz_r)
                Ey_ic = self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TE_Ey_r) * self.kux
                self.Bz_s[self.i_re_TE] = - self._dt / self._st.dx * Ey_ic  # - sign?
                self.Dy_s[self.i_re_TE] = - self._dt / self._st.dx * Hz_ic

            # ======  method 4  ======  todo: why doesn't this work?
            # # try directly calculate B and D fields
            # Hz_ic = [1.2 * self._amp * self._g.f(t - td) for td in [self.td_TE_Hz_t, self.td_TE_Hz_b, self.td_TE_Hz_l, self.td_TE_Hz_r]]
            # Ex_ic = [self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - td) * self.kuy for td in [self.td_TE_Ex_t, self.td_TE_Ex_b]]
            # Ey_ic = [self._amp * np.sqrt(self._mu / self._epsi) * self._g.f(t - td) * self.kux for td in [self.td_TE_Ey_l, self.td_TE_Ey_r]]
            # self.Bz_s[self.i_te_TE] = - self._mu * Hz_ic[0]
            # self.Bz_s[self.i_be_TE] = - self._mu * Hz_ic[1]
            # self.Bz_s[self.i_le_TE] = - self._mu * Hz_ic[2]
            # self.Bz_s[self.i_re_TE] = - self._mu * Hz_ic[3]
            # #
            # self.Dx_s[self.i_te_TE] = - self._epsi * Ex_ic[0]
            # self.Dx_s[self.i_be_TE] = - self._epsi * Ex_ic[1]
            # #
            # self.Dy_s[self.i_le_TE] = - self._epsi * Ey_ic[0]
            # self.Dy_s[self.i_re_TE] = - self._epsi * Ey_ic[1]

        elif self._plrz == 'Ez':
            # ======  method 3  ======
            # t1 = time.perf_counter()
            if self.tblr[0]:
                Ez_ic = self._amp * self._g.f(t - self.td_TM_Ez_t)
                Hx_ic = self._amp / np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TM_Hx_t) * self.kuy
                Hx_ic_tr_co = Hx_ic[-1]
                self.Dz_s[self.i_te_TM] = self._dt / self._st.dy * Hx_ic
                self.Bx_s[self.i_te_TM] = self._dt / self._st.dy * Ez_ic
            if self.tblr[1]:
                Ez_ic = self._amp * self._g.f(t - self.td_TM_Ez_b)
                Hx_ic = self._amp / np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TM_Hx_b) * self.kuy
                self.Dz_s[self.i_be_TM] = -self._dt / self._st.dy * Hx_ic
                self.Bx_s[self.i_be_TM] = -self._dt / self._st.dy * Ez_ic
            if self.tblr[2]:
                Ez_ic = self._amp * self._g.f(t - self.td_TM_Ez_l)
                Hy_ic = self._amp / np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TM_Hy_l) * self.kux
                self.Dz_s[self.i_le_TM] = -self._dt / self._st.dx * Hy_ic  # - sign?
                self.By_s[self.i_le_TM] = self._dt / self._st.dx * Ez_ic
            if self.tblr[3]:
                Ez_ic = self._amp * self._g.f(t - self.td_TM_Ez_r)
                Hy_ic = self._amp / np.sqrt(self._mu / self._epsi) * self._g.f(t - self.td_TM_Hy_r) * self.kux
                self.Dz_s[self.i_re_TM] = self._dt / self._st.dx * Hy_ic  # - sign?
                self.By_s[self.i_re_TM] = -self._dt / self._st.dx * Ez_ic
                if self.tblr[0]:
                    self.Dz_s[self.i_co_tr_TM] += self._dt / self._st.dy * Hx_ic_tr_co
            # t2 = time.perf_counter()
            # print("{:e}".format(t2-t1))



# -*- coding: utf-8 -*-


import time

import numpy as np
from matplotlib import pyplot as plt
# from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec
from matplotlib.widgets import Button, CheckButtons, TextBox
from matplotlib.animation import FFMpegWriter

from tdyno.dom import DOM
from tdyno.a_pchs import a_pchs
from tdyno.plot_field_2d_in_ax_3 import PlotField2DinAx
from tdyno.s2t import S2T
from tdyno.sc2t import PS2, TSS2
from tdyno.pct import PCT
from tdyno.mnt.mnt_t_fa import MntPntAmp, MntMltPntAmp
from tdyno.mnt.mnt_t_poyn import Mnt2DSqPoynU


class RT2:

    def __init__(self,
                 c0, epsi0, mu0,
                 st,
                 scs,
                 dt, Nt,
                 pc=None,
                 plrz='Hz',
                 if_ndc=False, omg_ndc=None,
                 md='HDE',
                 mnts=None,
                 skp=10,
                 **kwargs
                 ):

        """
        Run FDTD 2D simulation.

        Parameters
        ----------
        c0                                  :   float
        epsi0                               :   float
        mu0                                 :   float
        st                                  :   S2T
                                                the structure to be solved
        scs                                 :   list[PS2 | TSS2]
                                                sources
        dt                                  :   float
        Nt                                  :   int
                                                total number of time steps after each time you click Play/Pause button.
        pc                                  ：  PCT
                                                pml
        plrz                                :   str
                                                polarization, either 'Hz' or 'Ez'
        if_ndc                              :   bool
                                                whether to implement numeric dispersion compensation
        omg_ndc                             :   float
                                                the frequency for numeric dispersion compensation.
        md                                  :   str
                                                mode, either 'HDE' or 'BHDE'
        mnts                                :   list[MntMltPntAmp | MntPntAmp | Mnt2DSqPoynU]
                                                list of monitors
        skp                                 :   int
                                                skipping in plotting.

        Keyword arguments
        -----------------
        kwargs                              :   dict
                                                key word arguments for PlotField2DinAx

        Returns
        -------
        object


        Attributes
        ----------

        """

        self.c0 = c0
        self.epsi0 = epsi0
        self.mu0 = mu0
        self.st = st
        self.scs = scs
        self.pc = pc
        self.plrz = plrz
        self.md = md
        self.dt = dt
        self.Nt = Nt
        self.mnts = mnts
        self.skp = skp

        if not self.pc:
            self.if_pc = False
        else:
            self.if_pc = True

        # For dynamic modulation
        if self.st.bxs_dmri:
            self.epsi_x_dm = self.st.epsi_bg[0, 0] * np.ones([self.st.Ny, self.st.Nx])
            self.epsi_y_dm = self.st.epsi_bg[1, 1] * np.ones([self.st.Ny, self.st.Nx])
            self.epsi_z_dm = self.st.epsi_bg[2, 2] * np.ones([self.st.Ny, self.st.Nx])

        # For loss/gain boxes in structure
        self.if_lg = False
        if self.st.bxs_lg or self.st.bxs_dmlg:
            self.if_lg = True

            # index of the region not in loss/gain boxes.
            self.i_not_lg = np.ones([self.st.Ny, self.st.Nx], dtype=bool)

            # todo: This does not seem necessary. Doesn't need to store sgm in self.
            self.sgm_x = np.zeros([self.st.Ny, self.st.Nx])
            self.sgm_y = np.zeros([self.st.Ny, self.st.Nx])
            self.sgm_z = np.zeros([self.st.Ny, self.st.Nx])
            for box in (self.st.bxs_lg+self.st.bxs_dmlg):
                self.i_not_lg[box.gmt.iib] = False
                self.sgm_x[box.gmt.iib] = box.mtr.sgm
                self.sgm_y[box.gmt.iib] = box.mtr.sgm
                self.sgm_z[box.gmt.iib] = box.mtr.sgm
            self.i_not_lg_V = self.i_not_lg.ravel()
            self.sgm_x_V = self.sgm_x.ravel()
            self.sgm_y_V = self.sgm_y.ravel()
            self.sgm_z_V = self.sgm_z.ravel()

        self.t = 0
        self.psd = True  # paused

        # Initialize B, H, D, E fields
        self.Bx, self.By, self.Bz, self.Hx, self.Hy, self.Hz, self.Dx, self.Dy, self.Dz, self.Ex, self.Ey, self.Ez = [np.zeros([self.st.Ny, self.st.Nx]) for ii in range(12)]
        self.Bx_V, self.By_V, self.Bz_V, self.Hx_V, self.Hy_V, self.Hz_V, self.Dx_V, self.Dy_V, self.Dz_V, self.Ex_V, self.Ey_V, self.Ez_V = [f.ravel() for f in [self.Bx, self.By, self.Bz, self.Hx, self.Hy, self.Hz, self.Dx, self.Dy, self.Dz, self.Ex, self.Ey, self.Ez]]

        # Initialize psi, JP and P fields
        if self.pc:
            self.psi_Bzx_V_iP, self.psi_Bzy_V_iP, self.psi_Dxy_V_iP, self.psi_Dyx_V_iP = [np.zeros(self.pc.iiPV[self.pc.iiPV].size) for ii in range(4)]
            self.psi_Bxy_V_iP, self.psi_Byx_V_iP, self.psi_Dzx_V_iP, self.psi_Dzy_V_iP = [np.zeros(self.pc.iiPV[self.pc.iiPV].size) for ii in range(4)]
        self.JPx_Lz, self.JPy_Lz, self.JPz_Lz = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_Lz] for ii in range(3)]
        self.Px_Lz, self.Py_Lz, self.Pz_Lz = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_Lz] for ii in range(3)]
        self.JPx_Dr, self.JPy_Dr, self.JPz_Dr = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_Dr] for ii in range(3)]
        self.Px_Dr, self.Py_Dr, self.Pz_Dr = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_Dr] for ii in range(3)]
        self.JPx_dmLz, self.JPy_dmLz, self.JPz_dmLz = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_dmLz] for ii in range(3)]
        self.Px_dmLz, self.Py_dmLz, self.Pz_dmLz = [[[np.zeros(box.gmt.iib[box.gmt.iib].shape, dtype=float) for g in box.mtr.Gms] for box in self.st.bxs_dmLz] for ii in range(3)]

        #  Differential operators
        self.d = DOM(Nx=self.st.Nx, Ny=self.st.Ny, dx=self.st.dx, dy=self.st.dy)

        # numeric dispersion compensation
        self.set_ndc(if_ndc, omg_ndc)

        # ==========   Start the plotting window, handle plotting and user interface   ==========

        self.fig = plt.figure()
        gs = GridSpec(1, 1, top=0.85, bottom=0.25, left=0.1, right=0.9)

        # play/pause button
        self.ax_pp = plt.axes([0.1, 0.04, 0.1, 0.05])
        self.b_pp = Button(self.ax_pp, r'Play/Pause')
        self.b_pp.on_clicked(self.pp)

        # step button
        self.ax_s = plt.axes([0.25, 0.04, 0.1, 0.05])
        self.b_s = Button(self.ax_s, r'Step')
        self.b_s.on_clicked(self.stp)

        # pcm render button
        self.ax_pr = plt.axes([0.4, 0.04, 0.1, 0.05])
        self.b_pr = Button(self.ax_pr, r'Render')
        self.b_pr.on_clicked(self.pcm_rndr)
        self.fig_pr_p = None
        self.ax_pr_p = None

        # reset button
        self.ax_r = plt.axes([0.55, 0.04, 0.1, 0.05])
        self.b_r = Button(self.ax_r, 'Reset')
        self.b_r.on_clicked(self.r_f)

        # press 'right' key  = step
        self.fig.canvas.mpl_connect('key_press_event', self.k_p)

        # main ax containing the field plot
        self.ax = self.fig.add_subplot(gs[0, 0])  # Axes
        # if use blitting, need to set_animated
        # self.ax.set_animated(True)
        if plrz == 'Hz':
            sf = self.Hz
            title = 'Hz'
        else:  # i.e. plrz is 'Ez'
            sf = self.Ez
            title = 'Ez'

        # patch collection list
        self.phcs = []

        # collect different type of boxes (different edge colors)
        for box in self.st.bxs_sp:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[0., 0., 0., 0.5])  # black
            self.phcs.append(phc)
        for box in self.st.bxs_Lz:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[0., 0.3, 0.8, 1.])  # blue
            self.phcs.append(phc)
        for box in self.st.bxs_dmLz:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[1., 0.6, 0.1, 1.])  # orange
            self.phcs.append(phc)
        for box in self.st.bxs_Dr:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[1., 0.8, 0.2, 1.])  # yellow
            self.phcs.append(phc)
        for box in self.st.bxs_dmri:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[1., 0.2, 1., 1.])  # purple
            self.phcs.append(phc)
        for box in self.st.bxs_lg:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[0.1, 1., 0.2, 1.])  # green
            self.phcs.append(phc)
        for box in self.st.bxs_dmlg:
            phc = a_pchs(self.ax, [box.gmt.pch], fc='None', alpha=None, lw=0.75, ec=[0.7, 0.0, 0.0, 1.])  # red
            self.phcs.append(phc)

        # store the source box
        for sc in self.scs:
            if hasattr(sc, 'pchs') and sc.pchs:
                phc = a_pchs(self.ax, sc.pchs, fc='none', lw=0.75, alpha=0.4, ec='purple')
                self.phcs.append(phc)

        # store the monitor as a box
        if mnts is not None and mnts:
            for mnt in mnts:
                phc = a_pchs(self.ax, mnt.pchs, fc='none', lw=0.75, alpha=0.4, ec='black')
                self.phcs.append(phc)

        if self.if_pc:
            shaded_region = self.pc.iiP[:, :, 0]
        else:
            shaded_region = None
        self.plotter = PlotField2DinAx(
            self.ax,
            xmin=self.st.xmin, xmax=self.st.xmax, xres=self.st.dx, ymin=self.st.ymin, ymax=self.st.ymax, yres=self.st.dy, dire='v',
            title=title, xlabel='$z$', ylabel='$x$',
            sf=sf, if_scalar_field_intensity=True, scalar_field_part='real', if_colorbar=True,
            shaded_region=shaded_region, if_update_shading=False,
            **kwargs)

        # Show time steps
        bbox_props = dict(boxstyle="round, pad=0.3", fc=[.85, .9, 1., .9], lw=0)
        self.an_t = self.ax.annotate('time step : {:d}'.format(self.t), fontsize=14, xy=(0.5, 0.5), xycoords="data",
                                     xytext=(0.8, 0.065), textcoords='figure fraction',
                                     horizontalalignment='center', verticalalignment='center',
                                     bbox=bbox_props
                                     )

        # video recording parameters
        self.if_vid_rec = False
        self.if_vid_rec_ps = False
        self.vid_fps = 30
        self.vid_dpi = 100
        self.vid_n = 0
        self.vid_ttl = 'v'
        self.vid_of = self.vid_ttl + '_{:d}.mp4'.format(self.vid_n)
        self.vid_metadata = dict(artist='Alex Y. Song')

        # video writer
        self.vid_wtr = FFMpegWriter(fps=self.vid_fps, metadata=self.vid_metadata)
        # self.vid_wtr.setup(fig=self.fig, outfile=self.vid_of)

        # video recording button
        self.ax_vid_rec = plt.axes([0.1, 0.93, 0.085, 0.05])
        self.b_vid_rec = CheckButtons(self.ax_vid_rec, [r'Record'], [False])
        self.b_vid_rec.on_clicked(self.vid_rec)

        # video recording pause
        self.ax_vid_rec_ps = plt.axes([0.195, 0.93, 0.08, 0.05])
        self.b_vid_rec_ps = CheckButtons(self.ax_vid_rec_ps, [r'Pause'], [False])
        self.b_vid_rec_ps.on_clicked(self.vid_rec_ps)

        # video fps
        self.ax_vid_fps = plt.axes([0.31, 0.93, 0.04, 0.05])
        self.tb_vid_fps = TextBox(self.ax_vid_fps, 'fps', initial='{:d}'.format(self.vid_fps))
        self.tb_vid_fps.on_submit(self.set_vid_fps)

        # video dpi
        self.ax_vid_dpi = plt.axes([0.39, 0.93, 0.05, 0.05])
        self.tb_vid_dpi = TextBox(self.ax_vid_dpi, 'dpi', initial='{:d}'.format(self.vid_dpi))
        self.tb_vid_dpi.on_submit(self.set_vid_dpi)

        # video title
        self.ax_vid_ttl = plt.axes([0.54, 0.93, 0.35, 0.05])
        self.tb_vid_ttl = TextBox(self.ax_vid_ttl, 'file name', initial=self.vid_ttl)
        self.tb_vid_ttl.on_text_change(self.set_vid_ttl)

        self.fig.set_size_inches(8., 5.01)

        # save the ax for blitting
        # self.ax_cache = self.fig.canvas.copy_from_bbox(self.ax.bbox)
        # self.fig.draw_artist(self.ax)

    def set_ndc(self, if_ndc, omg_ndc):
        """
        set numeric dispersion correction

        Parameters
        ----------
        if_ndc      :   bool
        omg_ndc     :   float

        Returns
        -------

        """
        c0 = self.c0
        dt = self.dt

        if omg_ndc is None:
            omg_ndc = c0 / .2
        self.omg_ndc = omg_ndc
        if if_ndc:
            n = np.sqrt(self.st.epsi_bg[2, 2] * self.st.mu_bg[2, 2])  # todo: anisotropy case
            self.ndc_45deg = c0 * dt * np.sqrt(2.) / n / self.st.dx * np.sin(omg_ndc / 2. / np.sqrt(2.) * n * self.st.dx / c0) / np.sin(omg_ndc / 2. * dt)
            self.ndc_0deg = c0 * dt / n / self.st.dx * np.sin(omg_ndc / 2. * n * self.st.dx / c0) / np.sin(omg_ndc / 2. * dt)
            self.ndc = (self.ndc_0deg + self.ndc_45deg) / 2.
        else:
            self.ndc = 1.

    def vid_rec(self, event):
        """
        video recording.

        Parameters
        ----------
        event

        Returns
        -------

        """
        if self.if_vid_rec:
            self.vid_wtr.finish()
            self.vid_n += 1
        else:
            self.vid_of = self.vid_ttl + '_{:d}.mp4'.format(self.vid_n)
            self.vid_wtr.setup(fig=self.fig, outfile=self.vid_of, dpi=self.vid_dpi)
        self.if_vid_rec = self.b_vid_rec.get_status()[0]

    def vid_rec_ps(self, event):
        """
        If video recording is paused

        Parameters
        ----------
        event

        Returns
        -------

        """
        self.if_vid_rec_ps = self.b_vid_rec_ps.get_status()[0]

    def set_vid_fps(self, text):
        """
        set video fps

        Parameters
        ----------
        text

        Returns
        -------

        """
        self.vid_fps = eval(text)
        self.vid_wtr.fps = self.vid_fps

    def set_vid_dpi(self, text):
        """
        set the video dpi

        Parameters
        ----------
        text

        Returns
        -------

        """
        self.vid_dpi = eval(text)
        self.vid_wtr.dpi = self.vid_dpi

    def set_vid_ttl(self, text):
        """
        set recording video title

        Parameters
        ----------
        text

        Returns
        -------

        """
        self.vid_ttl = text
        self.vid_n = 0

    def pp(self, event):

        """
        play / pause.

        Parameters
        ----------
        event

        Returns
        -------

        """

        self.psd ^= True
        if not self.psd:
            self.rn()

    def stp(self, event):

        """
        Move forward one step in time.

        Parameters
        ----------
        event

        Returns
        -------

        """

        if self.psd:
            self.t += 1
            t1 = time.process_time()
            self.uf()
            t2 = time.process_time()
            if self.mnts is not None and self.mnts:
                if self.plrz == 'Hz':
                    for mnt in self.mnts:
                        mnt.rnf(self.Hz, self.Ex, self.Ey)
                else:
                    for mnt in self.mnts:
                        mnt.rnf(self.Ez, self.Hx, self.Hy)
            t3 = time.process_time()
            self.up()
            if self.mnts is not None and self.mnts:
                for mnt in self.mnts:
                    mnt.up()
            self.fig.canvas.flush_events()
            t4 = time.process_time()
            if self.if_vid_rec and not self.if_vid_rec_ps:
                self.vid_wtr.grab_frame()
            t5 = time.process_time()
            print('Computation time: {:.5f} | Monitor time: {:.5f} | Plotting time: {:.5f}'.format((t2 - t1), (t3 - t2), (t4 - t3)))

    def k_p(self, event):

        """
        key pressed.

        Parameters
        ----------
        event

        Returns
        -------

        """

        if event.key == 'right':
            self.stp(event)
        elif event.key == ' ':
            self.pp(None)

    def pcm_rndr(self, event):
        """
        pcolormesh render

        Parameters
        ----------
        event

        Returns
        -------

        """
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_aspect('equal')
        if self.plrz == 'Hz':
            self.plotter.plot_sf_intensity_temp_pcm(ax=ax, sf=self.Hz)
        elif self.plrz == 'Ez':
            self.plotter.plot_sf_intensity_temp_pcm(ax=ax, sf=self.Ez)
        else:
            print('What polarization is this?')

        # self.fig.canvas.update()
        plt.show(block=False)

        self.fig_pr_p = fig
        self.ax_pr_p = ax

    def r_f(self, event):

        """
        reset fields

        Parameters
        ----------
        event

        Returns
        -------

        """

        for f in [self.Bx, self.By, self.Bz, self. Hx, self.Hy, self.Hz, self.Dx, self.Dy, self.Dz, self.Ex, self.Ey, self.Ez]:
            f.fill(0.)

        for f in [self.psi_Bzx_V_iP, self.psi_Bzy_V_iP, self.psi_Dxy_V_iP, self.psi_Dyx_V_iP, self.psi_Dzx_V_iP, self.psi_Dzy_V_iP, self.psi_Bxy_V_iP, self.psi_Byx_V_iP]:
            f.fill(0.)

        for JP in [self.JPx_Lz, self.JPy_Lz, self.JPz_Lz]:
            for JP_box in JP:
                for jp in JP_box:
                    jp.fill(0.)
        for P in [self.Px_Lz, self.Py_Lz, self.Pz_Lz]:
            for P_box in P:
                for p in P_box:
                    p.fill(0.)
        for JP in [self.JPx_Dr, self.JPy_Dr, self.JPz_Dr]:
            for JP_box in JP:
                for jp in JP_box:
                    jp.fill(0.)
        for P in [self.Px_Dr, self.Py_Dr, self.Pz_Dr]:
            for P_box in P:
                for p in P_box:
                    p.fill(0.)
        for JP in [self.JPx_dmLz, self.JPy_dmLz, self.JPz_dmLz]:
            for JP_box in JP:
                for jp in JP_box:
                    jp.fill(0.)
        for P in [self.Px_dmLz, self.Py_dmLz, self.Pz_dmLz]:
            for P_box in P:
                for p in P_box:
                    p.fill(0.)
        self.t = 0
        self.up()
        if self.mnts is not None and self.mnts:
            for mnt in self.mnts:
                mnt.cf()

    def rn(self):

        """
        running simulation

        Returns
        -------

        """

        ttt = 0
        while ttt < self.Nt:
            # print(self.t * self.dt)
            if not self.psd:
                t1 = time.perf_counter()
                self.uf()
                t2 = time.perf_counter()
                if self.mnts is not None and self.mnts:
                    if self.plrz == 'Hz':
                        for mnt in self.mnts:
                            mnt.rnf(self.Hz, self.Ex, self.Ey)
                    else:
                        for mnt in self.mnts:
                            mnt.rnf(self.Ez, self.Hx, self.Hy)
                t3 = time.perf_counter()
                if self.t % self.skp == 0:
                    self.up()
                    if self.mnts is not None and self.mnts:
                        for mnt in self.mnts:
                            mnt.up()
                    # plt.pause(0.0001)
                    self.fig.canvas.flush_events()
                t4 = time.perf_counter()
                if self.t % self.skp == 0:
                    if self.if_vid_rec and not self.if_vid_rec_ps:
                        self.vid_wtr.grab_frame()
                t5 = time.perf_counter()
                print('Computation time: {:.5f} | Monitor time: {:.5f} | Plotting time: {:.5f}'.format((t2-t1), (t3-t2), (t4 - t3)))
                ttt += 1
                self.t += 1
            else:
                ttt = self.Nt
                print('paused!')

        self.psd = True

    def uf(self):

        """
        update fields.

        Returns
        -------

        """

        if self.if_pc:  # has PML
            iiP = self.pc.iiPV

            if self.plrz == 'Hz':

                # PML psi fields
                self.psi_Bzx_V_iP = self.pc.bxfViP * self.psi_Bzx_V_iP + self.pc.cxfViP * (self.d.pxf * self.Ey_V)[iiP]
                self.psi_Bzy_V_iP = self.pc.byfViP * self.psi_Bzy_V_iP + self.pc.cyfViP * (self.d.pyf * self.Ex_V)[iiP]

                # update B fields
                self.Bz_V += - self.c0 * self.dt * (self.pc.kixV * (self.d.pxf * self.Ey_V) - self.pc.kiyV * (self.d.pyf * self.Ex_V))
                self.Bz_V[iiP] += - self.c0 * self.dt * (self.psi_Bzx_V_iP - self.psi_Bzy_V_iP)

                # add source to B fields
                for sc in self.scs:
                    sc.us(self.dt * self.t)
                    self.Bz[sc.iB_s] += sc.Bz_s[sc.iB_s]

                # update H fields
                self.Hz = 1. / self.st.mu_bg[2, 2] / self.ndc * self.Bz
                if self.md is 'BHDE':
                    for bx in self.st.bxs:
                        self.Hz[bx.gmt.iib] = 1. / bx.mtr.mu[2, 2] / self.ndc * self.Bz[bx.gmt.iib]
                self.Hz_V = self.Hz.reshape(self.st.Ny * self.st.Nx)

                # Psi fields
                self.psi_Dxy_V_iP = self.pc.bybViP * self.psi_Dxy_V_iP + self.pc.cybViP * (self.d.pyb * self.Hz_V)[iiP]
                self.psi_Dyx_V_iP = self.pc.bxbViP * self.psi_Dyx_V_iP + self.pc.cxbViP * (self.d.pxb * self.Hz_V)[iiP]

                # update D fields
                if not self.if_lg:
                    self.Dx_V += self.c0 * self.dt * (self.pc.kiyV * (self.d.pyb * self.Hz_V))
                    self.Dy_V += - self.c0 * self.dt * (self.pc.kixV * (self.d.pxb * self.Hz_V))
                else:
                    self.Dx_V[self.i_not_lg_V] += self.c0 * self.dt * (self.pc.kiyV * (self.d.pyb * self.Hz_V))[self.i_not_lg_V]
                    self.Dy_V[self.i_not_lg_V] += - self.c0 * self.dt * (self.pc.kixV * (self.d.pxb * self.Hz_V))[self.i_not_lg_V]
                    # loss/gain
                    for bx in self.st.bxs_lg:
                        Dx_V_ib_pv = (1./self.dt - bx.mtr.sgm/2./bx.mtr.epsi[2, 2]) * self.Dx_V[bx.gmt.iibV]
                        self.Dx_V[bx.gmt.iibV] = (1./(1./self.dt + bx.mtr.sgm/2./bx.mtr.epsi[2, 2])) * (self.c0 * (self.pc.kiyV[bx.gmt.iibV] * (self.d.pyb * self.Hz_V)[bx.gmt.iibV]) + Dx_V_ib_pv)
                        Dy_V_ib_pv = (1./self.dt - bx.mtr.sgm/2./bx.mtr.epsi[2, 2]) * self.Dy_V[bx.gmt.iibV]
                        self.Dy_V[bx.gmt.iibV] = (1./(1./self.dt + bx.mtr.sgm/2./bx.mtr.epsi[2, 2])) * ((-self.c0) * (self.pc.kixV[bx.gmt.iibV] * (self.d.pxb * self.Hz_V)[bx.gmt.iibV]) + Dy_V_ib_pv)
                    # dynamic loss/gain modulations
                    for bx in self.st.bxs_dmlg:
                        self.sgm_x[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        self.sgm_y[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        Dx_V_ib_pv = (1. / self.dt - self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dx_V[bx.gmt.iibV]
                        self.Dx_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * (self.c0 * (self.pc.kiyV[bx.gmt.iibV] * (self.d.pyb * self.Hz_V)[bx.gmt.iibV]) + Dx_V_ib_pv)
                        Dy_V_ib_pv = (1. / self.dt - self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dy_V[bx.gmt.iibV]
                        self.Dy_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * ((-self.c0) * (self.pc.kixV[bx.gmt.iibV] * (self.d.pxb * self.Hz_V)[bx.gmt.iibV]) + Dy_V_ib_pv)

                # handle PML
                self.Dx_V[iiP] += self.c0 * self.dt * self.psi_Dxy_V_iP
                self.Dy_V[iiP] += - self.c0 * self.dt * self.psi_Dyx_V_iP

                # add source to D fields
                for sc in self.scs:
                    self.Dx[sc.iD_s] += sc.Dx_s[sc.iD_s]
                    self.Dy[sc.iD_s] += sc.Dy_s[sc.iD_s]

                # JP and P fields
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_Lz, self.JPy_Lz, self.Px_Lz, self.Py_Lz, self.st.bxs_Lz):
                    for JPx, JPy, Px, Py, omg, dlt_epsi, Gm in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms):
                        JPx[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPx - omg**2 * Px + dlt_epsi * omg**2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPy - omg**2 * Py + dlt_epsi * omg**2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_Dr, self.JPy_Dr, self.Px_Dr, self.Py_Dr, self.st.bxs_Dr):
                    for JPx, JPy, Px, Py, omg, Gm in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.Gms):
                        JPx[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPx + omg**2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPy + omg**2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_dmLz, self.JPy_dmLz, self.Px_dmLz, self.Py_dmLz, self.st.bxs_dmLz):
                    for JPx, JPy, Px, Py, omg, dlt_epsi, Gm, m_a, m_o, m_q, m_p in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms, bx.mtr.m_a, bx.mtr.m_o, bx.mtr.m_q, bx.mtr.m_p):
                        dlt_epsi_a = dlt_epsi + m_a * np.cos(-m_o * (self.t - 1. / 2.) * self.dt - m_q[0] * self.st.xx[bx.gmt.iib] - m_q[1] * self.st.yy[bx.gmt.iib] + m_p)
                        JPx[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPx - omg**2 * Px + dlt_epsi_a * omg**2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPy - omg**2 * Py + dlt_epsi_a * omg**2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy

                # update E fields
                self.Ex = 1. / self.st.epsi_bg[0, 0] / self.ndc * self.Dx
                self.Ey = 1. / self.st.epsi_bg[1, 1] / self.ndc * self.Dy
                # simple boxes
                for bx in self.st.bxs_sp:
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib])
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib])
                for bx, Px_b, Py_b in zip((self.st.bxs_Lz+self.st.bxs_Dr+self.st.bxs_dmLz), (self.Px_Lz+self.Px_Dr+self.Px_dmLz), (self.Py_Lz+self.Py_Dr+self.Py_dmLz)):
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib] - sum([p for p in Px_b]))
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib] - sum([p for p in Py_b]))
                for bx in (self.st.bxs_lg + self.st.bxs_dmlg):
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib])
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib])
                # dynamic modulations
                for bx in self.st.bxs_dmri:
                    self.epsi_x_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    # todo: why all negative sign?
                    self.epsi_y_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    self.Ey[bx.gmt.iib] = 1. / self.epsi_y_dm[bx.gmt.iib] / self.ndc * self.Dy[bx.gmt.iib]

                self.Ex_V = self.Ex.reshape(self.st.Ny * self.st.Nx)
                self.Ey_V = self.Ey.reshape(self.st.Ny * self.st.Nx)

            elif self.plrz == 'Ez':

                # Psi fields
                self.psi_Bxy_V_iP = self.pc.byfViP * self.psi_Bxy_V_iP + self.pc.cyfViP * (self.d.pyf * self.Ez_V)[iiP]
                self.psi_Byx_V_iP = self.pc.bxfViP * self.psi_Byx_V_iP + self.pc.cxfViP * (self.d.pxf * self.Ez_V)[iiP]

                # update B fields
                self.Bx_V += - self.c0 * self.dt * (self.pc.kiyV * (self.d.pyf * self.Ez_V))
                self.Bx_V[iiP] += - self.c0 * self.dt * self.psi_Bxy_V_iP
                self.By_V += self.c0 * self.dt * (self.pc.kixV * (self.d.pxf * self.Ez_V))
                self.By_V[iiP] += self.c0 * self.dt * self.psi_Byx_V_iP

                # add source to B fields
                for sc in self.scs:
                    sc.us(self.dt * self.t)
                    self.Bx[sc.iB_s] += sc.Bx_s[sc.iB_s]
                    self.By[sc.iB_s] += sc.By_s[sc.iB_s]

                # update H fields
                self.Hx = 1. / self.st.mu_bg[0, 0] / self.ndc * self.Bx
                self.Hy = 1. / self.st.mu_bg[1, 1] / self.ndc * self.By
                if self.md is 'BHDE':
                    for bx in self.st.bxs:
                        self.Hx[bx.gmt.iib] = 1. / bx.mtr.mu[0, 0] / self.ndc * self.Bx[bx.gmt.iib]
                        self.Hy[bx.gmt.iib] = 1. / bx.mtr.mu[0, 0] / self.ndc * self.By[bx.gmt.iib]
                self.Hx_V = self.Hx.reshape(self.st.Ny * self.st.Nx)
                self.Hy_V = self.Hy.reshape(self.st.Ny * self.st.Nx)

                # Psi fields
                self.psi_Dzx_V_iP = self.pc.bxbViP * self.psi_Dzx_V_iP + self.pc.cxbViP * (self.d.pxb * self.Hy_V)[iiP]
                self.psi_Dzy_V_iP = self.pc.bybViP * self.psi_Dzy_V_iP + self.pc.cybViP * (self.d.pyb * self.Hx_V)[iiP]

                # update D fields
                if not self.if_lg:
                    self.Dz_V += self.c0 * self.dt * (self.pc.kixV * (self.d.pxb * self.Hy_V) - self.pc.kiyV * (self.d.pyb * self.Hx_V))
                else:
                    self.Dz_V[self.i_not_lg_V] += self.c0 * self.dt * (self.pc.kixV * (self.d.pxb * self.Hy_V) - self.pc.kiyV * (self.d.pyb * self.Hx_V))[self.i_not_lg_V]
                    # loss/gain
                    for bx in self.st.bxs_lg:
                        Dz_V_ib_pv = (1./self.dt - bx.mtr.sgm/2./bx.mtr.epsi[2, 2]) * self.Dz_V[bx.gmt.iibV]
                        self.Dz_V[bx.gmt.iibV] = (1. / (1./self.dt + bx.mtr.sgm/2./bx.mtr.epsi[2, 2])) * (self.c0 * (self.pc.kixV[bx.gmt.iibV] * (self.d.pxb * self.Hy_V)[bx.gmt.iibV] - self.pc.kiyV[bx.gmt.iibV] * (self.d.pyb * self.Hx_V)[bx.gmt.iibV]) + Dz_V_ib_pv)
                    # dynamic gain/loss modulations
                    for bx in self.st.bxs_dmlg:
                        self.sgm_z[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        Dz_V_ib_pv = (1. / self.dt - self.sgm_z_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dz_V[bx.gmt.iibV]
                        self.Dz_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_z_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * (self.c0 * (self.pc.kixV[bx.gmt.iibV] * (self.d.pxb * self.Hy_V)[bx.gmt.iibV] - self.pc.kiyV[bx.gmt.iibV] * (self.d.pyb * self.Hx_V)[bx.gmt.iibV]) + Dz_V_ib_pv)

                # handle PML
                self.Dz_V[iiP] += self.c0 * self.dt * (self.psi_Dzx_V_iP - self.psi_Dzy_V_iP)

                # add source to D fields
                for sc in self.scs:
                    self.Dz[sc.iD_s] += sc.Dz_s[sc.iD_s]

                # JP and P fields
                for JPz_b, Pz_b, bx in zip(self.JPz_Lz, self.Pz_Lz, self.st.bxs_Lz):
                    for JPz, Pz, omg, dlt_epsi, Gm in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms):
                        JPz[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPz - omg**2 * Pz + dlt_epsi * omg**2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz
                for JPz_b, Pz_b, bx in zip(self.JPz_Dr, self.Pz_Dr, self.st.bxs_Dr):
                    for JPz, Pz, omg, Gm in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.Gms):
                        JPz[:] = 1. / (1./self.dt + Gm/2.) * ((1./self.dt - Gm/2.) * JPz + omg**2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz
                for JPz_b, Pz_b, bx in zip(self.JPz_dmLz, self.Pz_dmLz, self.st.bxs_dmLz):
                    for JPz, Pz, omg, dlt_epsi, Gm, m_a, m_o, m_q, m_p in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms, bx.mtr.m_a, bx.mtr.m_o, bx.mtr.m_q, bx.mtr.m_p):
                        dlt_epsi_a = dlt_epsi + m_a * np.cos(-m_o * (self.t - 1. / 2.) * self.dt - m_q[0] * self.st.xx[bx.gmt.iib] - m_q[1] * self.st.yy[bx.gmt.iib] + m_p)
                        JPz[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPz - omg ** 2 * Pz + dlt_epsi_a * omg ** 2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz

                # update E fields
                self.Ez = 1. / self.st.epsi_bg[2, 2] / self.ndc * self.Dz
                # simple boxes
                for bx in (self.st.bxs_sp):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib])
                for bx, Pz_b, in zip((self.st.bxs_Lz+self.st.bxs_Dr+self.st.bxs_dmLz), (self.Pz_Lz + self.Pz_Dr + self.Pz_dmLz)):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib] - sum([p for p in Pz_b]))
                for bx in (self.st.bxs_lg + self.st.bxs_dmlg):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib])
                # dynamic modulations
                for bx in self.st.bxs_dmri:
                    self.epsi_z_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    self.Ez[bx.gmt.iib] = 1. / self.epsi_z_dm[bx.gmt.iib] / self.ndc * self.Dz[bx.gmt.iib]

                self.Ez_V = self.Ez.ravel()

        else:  # no PML

            if self.plrz == 'Hz':

                # update B fields
                self.Bz_V += - self.c0 * self.dt * ((self.d.pxf * self.Ey_V) - (self.d.pyf * self.Ex_V))

                # add source to B fields
                for sc in self.scs:
                    sc.us(self.dt * self.t)
                    self.Bz[sc.iB_s] += sc.Bz_s[sc.iB_s]

                # update H fields
                self.Hz = 1. / self.st.mu_bg[2, 2] / self.ndc * self.Bz
                if self.md is 'BHDE':
                    for bx in self.st.bxs:
                        self.Hz[bx.gmt.iib] = 1. / bx.mtr.mu[2, 2] / self.ndc * self.Bz[bx.gmt.iib]
                self.Hz_V = self.Hz.reshape(self.st.Ny * self.st.Nx)

                # update D fields
                if not self.if_lg:
                    self.Dx_V += self.c0 * self.dt * ((self.d.pyb * self.Hz_V))
                    self.Dy_V += - self.c0 * self.dt * ((self.d.pxb * self.Hz_V))
                else:
                    self.Dx_V[self.i_not_lg_V] += self.c0 * self.dt * ((self.d.pyb * self.Hz_V))[self.i_not_lg_V]
                    self.Dy_V[self.i_not_lg_V] += - self.c0 * self.dt * ((self.d.pxb * self.Hz_V))[self.i_not_lg_V]
                    # loss/gain
                    for bx in self.st.bxs_lg:
                        Dx_V_ib_pv = (1. / self.dt - bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2]) * self.Dx_V[bx.gmt.iibV]
                        self.Dx_V[bx.gmt.iibV] = (1. / (1. / self.dt + bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2])) * (self.c0 * ((self.d.pyb * self.Hz_V)[bx.gmt.iibV]) + Dx_V_ib_pv)
                        Dy_V_ib_pv = (1. / self.dt - bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2]) * self.Dy_V[bx.gmt.iibV]
                        self.Dy_V[bx.gmt.iibV] = (1. / (1. / self.dt + bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2])) * ((-self.c0) * ((self.d.pxb * self.Hz_V)[bx.gmt.iibV]) + Dy_V_ib_pv)
                    # dynamic loss/gain modulations
                    for bx in self.st.bxs_dmlg:
                        self.sgm_x[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        self.sgm_y[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        Dx_V_ib_pv = (1. / self.dt - self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dx_V[bx.gmt.iibV]
                        self.Dx_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * (self.c0 * ((self.d.pyb * self.Hz_V)[bx.gmt.iibV]) + Dx_V_ib_pv)
                        Dy_V_ib_pv = (1. / self.dt - self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dy_V[bx.gmt.iibV]
                        self.Dy_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_x_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * ((-self.c0) * ((self.d.pxb * self.Hz_V)[bx.gmt.iibV]) + Dy_V_ib_pv)

                # add source to D fields
                for sc in self.scs:
                    self.Dx[sc.iD_s] += sc.Dx_s[sc.iD_s]
                    self.Dy[sc.iD_s] += sc.Dy_s[sc.iD_s]

                # JP and P fields
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_Lz, self.JPy_Lz, self.Px_Lz, self.Py_Lz, self.st.bxs_Lz):
                    for JPx, JPy, Px, Py, omg, dlt_epsi, Gm in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms):
                        JPx[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPx - omg ** 2 * Px + dlt_epsi * omg ** 2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPy - omg ** 2 * Py + dlt_epsi * omg ** 2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_Dr, self.JPy_Dr, self.Px_Dr, self.Py_Dr, self.st.bxs_Dr):
                    for JPx, JPy, Px, Py, omg, Gm in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.Gms):
                        JPx[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPx + omg ** 2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPy + omg ** 2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy
                for JPx_b, JPy_b, Px_b, Py_b, bx in zip(self.JPx_dmLz, self.JPy_dmLz, self.Px_dmLz, self.Py_dmLz, self.st.bxs_dmLz):
                    for JPx, JPy, Px, Py, omg, dlt_epsi, Gm, m_a, m_o, m_q, m_p in zip(JPx_b, JPy_b, Px_b, Py_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms, bx.mtr.m_a, bx.mtr.m_o, bx.mtr.m_q, bx.mtr.m_p):
                        dlt_epsi_a = dlt_epsi + m_a * np.cos(-m_o * (self.t - 1. / 2.) * self.dt - m_q[0] * self.st.xx[bx.gmt.iib] - m_q[1] * self.st.yy[bx.gmt.iib] + m_p)
                        JPx[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPx - omg ** 2 * Px + dlt_epsi_a * omg ** 2 * self.Ex[bx.gmt.iib])
                        JPy[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPy - omg ** 2 * Py + dlt_epsi_a * omg ** 2 * self.Ey[bx.gmt.iib])
                        Px += self.dt * JPx
                        Py += self.dt * JPy

                # update E fields
                self.Ex = 1. / self.st.epsi_bg[0, 0] / self.ndc * self.Dx
                self.Ey = 1. / self.st.epsi_bg[1, 1] / self.ndc * self.Dy
                # simple boxes
                for bx in self.st.bxs_sp:
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib])
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib])
                for bx, Px_b, Py_b in zip((self.st.bxs_Lz + self.st.bxs_Dr + self.st.bxs_dmLz), (self.Px_Lz + self.Px_Dr + self.Px_dmLz), (self.Py_Lz + self.Py_Dr + self.Py_dmLz)):
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib] - sum([p for p in Px_b]))
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib] - sum([p for p in Py_b]))
                for bx in (self.st.bxs_lg + self.st.bxs_dmlg):
                    self.Ex[bx.gmt.iib] = 1. / bx.mtr.epsi[0, 0] / self.ndc * (self.Dx[bx.gmt.iib])
                    self.Ey[bx.gmt.iib] = 1. / bx.mtr.epsi[1, 1] / self.ndc * (self.Dy[bx.gmt.iib])
                # dynamic modulations
                for bx in self.st.bxs_dmri:
                    self.epsi_x_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    # todo: why all negative sign?
                    self.epsi_y_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    self.Ey[bx.gmt.iib] = 1. / self.epsi_y_dm[bx.gmt.iib] / self.ndc * self.Dy[bx.gmt.iib]

                self.Ex_V = self.Ex.reshape(self.st.Ny * self.st.Nx)
                self.Ey_V = self.Ey.reshape(self.st.Ny * self.st.Nx)

            elif self.plrz == 'Ez':

                # update B fields
                self.Bx_V += - self.c0 * self.dt * ((self.d.pyf * self.Ez_V))
                self.By_V += self.c0 * self.dt * ((self.d.pxf * self.Ez_V))

                # add source to B fields
                for sc in self.scs:
                    sc.us(self.dt * self.t)
                    self.Bx[sc.iB_s] += sc.Bx_s[sc.iB_s]
                    self.By[sc.iB_s] += sc.By_s[sc.iB_s]

                # update H fields
                self.Hx = 1. / self.st.mu_bg[0, 0] / self.ndc * self.Bx
                self.Hy = 1. / self.st.mu_bg[1, 1] / self.ndc * self.By
                if self.md is 'BHDE':
                    for bx in self.st.bxs:
                        self.Hx[bx.gmt.iib] = 1. / bx.mtr.mu[0, 0] / self.ndc * self.Bx[bx.gmt.iib]
                        self.Hy[bx.gmt.iib] = 1. / bx.mtr.mu[0, 0] / self.ndc * self.By[bx.gmt.iib]
                self.Hx_V = self.Hx.reshape(self.st.Ny * self.st.Nx)
                self.Hy_V = self.Hy.reshape(self.st.Ny * self.st.Nx)

                # update D fields
                if not self.if_lg:
                    self.Dz_V += self.c0 * self.dt * ((self.d.pxb * self.Hy_V) - (self.d.pyb * self.Hx_V))
                else:
                    self.Dz_V[self.i_not_lg_V] += self.c0 * self.dt * ((self.d.pxb * self.Hy_V) - (self.d.pyb * self.Hx_V))[self.i_not_lg_V]
                    # loss/gain
                    for bx in self.st.bxs_lg:
                        Dz_V_ib_pv = (1. / self.dt - bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2]) * self.Dz_V[bx.gmt.iibV]
                        self.Dz_V[bx.gmt.iibV] = (1. / (1. / self.dt + bx.mtr.sgm / 2. / bx.mtr.epsi[2, 2])) * (
                                    self.c0 * ((self.d.pxb * self.Hy_V)[bx.gmt.iibV] - (self.d.pyb * self.Hx_V)[bx.gmt.iibV]) + Dz_V_ib_pv)
                    # dynamic gain/loss modulations
                    for bx in self.st.bxs_dmlg:
                        self.sgm_z[bx.gmt.iib] = bx.mtr.sgm + bx.mtr.m_a * np.cos(-bx.mtr.m_o * (self.t - 1. / 2.) * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                        Dz_V_ib_pv = (1. / self.dt - self.sgm_z_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2]) * self.Dz_V[bx.gmt.iibV]
                        self.Dz_V[bx.gmt.iibV] = (1. / (1. / self.dt + self.sgm_z_V[bx.gmt.iibV] / 2. / bx.mtr.epsi[2, 2])) * (
                                    self.c0 * ((self.d.pxb * self.Hy_V)[bx.gmt.iibV] - (self.d.pyb * self.Hx_V)[bx.gmt.iibV]) + Dz_V_ib_pv)

                # add source to D fields
                for sc in self.scs:
                    self.Dz[sc.iD_s] += sc.Dz_s[sc.iD_s]

                # JP and P fields
                for JPz_b, Pz_b, bx in zip(self.JPz_Lz, self.Pz_Lz, self.st.bxs_Lz):
                    for JPz, Pz, omg, dlt_epsi, Gm in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms):
                        JPz[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPz - omg ** 2 * Pz + dlt_epsi * omg ** 2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz
                for JPz_b, Pz_b, bx in zip(self.JPz_Dr, self.Pz_Dr, self.st.bxs_Dr):
                    for JPz, Pz, omg, Gm in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.Gms):
                        JPz[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPz + omg ** 2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz
                for JPz_b, Pz_b, bx in zip(self.JPz_dmLz, self.Pz_dmLz, self.st.bxs_dmLz):
                    for JPz, Pz, omg, dlt_epsi, Gm, m_a, m_o, m_q, m_p in zip(JPz_b, Pz_b, bx.mtr.omgs_rsn, bx.mtr.dlts_epsi, bx.mtr.Gms, bx.mtr.m_a, bx.mtr.m_o, bx.mtr.m_q, bx.mtr.m_p):
                        dlt_epsi_a = dlt_epsi + m_a * np.cos(-m_o * (self.t - 1. / 2.) * self.dt - m_q[0] * self.st.xx[bx.gmt.iib] - m_q[1] * self.st.yy[bx.gmt.iib] + m_p)
                        JPz[:] = 1. / (1. / self.dt + Gm / 2.) * ((1. / self.dt - Gm / 2.) * JPz - omg ** 2 * Pz + dlt_epsi_a * omg ** 2 * self.Ez[bx.gmt.iib])
                        Pz += self.dt * JPz

                # update E fields
                self.Ez = 1. / self.st.epsi_bg[2, 2] / self.ndc * self.Dz
                # simple boxes
                for bx in (self.st.bxs_sp):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib])
                for bx, Pz_b, in zip((self.st.bxs_Lz + self.st.bxs_Dr + self.st.bxs_dmLz), (self.Pz_Lz + self.Pz_Dr + self.Pz_dmLz)):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib] - sum([p for p in Pz_b]))
                for bx in (self.st.bxs_lg + self.st.bxs_dmlg):
                    self.Ez[bx.gmt.iib] = 1. / bx.mtr.epsi[2, 2] / self.ndc * (self.Dz[bx.gmt.iib])
                # dynamic modulations
                for bx in self.st.bxs_dmri:
                    self.epsi_z_dm[bx.gmt.iib] = bx.mtr.epsi[2, 2] + bx.mtr.m_a * np.cos(-bx.mtr.m_o * self.t * self.dt - bx.mtr.m_q[0] * self.st.xx[bx.gmt.iib] - bx.mtr.m_q[1] * self.st.yy[bx.gmt.iib] + bx.mtr.m_p)
                    self.Ez[bx.gmt.iib] = 1. / self.epsi_z_dm[bx.gmt.iib] / self.ndc * self.Dz[bx.gmt.iib]

                self.Ez_V = self.Ez.ravel()

    def up(self):

        """
        update plots.

        Returns
        -------

        """

        # for blitting, first restore
        # self.ax.figure.canvas.restore_region(self.ax_cache)

        if self.plrz == 'Hz':
            # self.sf_pc.set_array(self.Hz)
            self.plotter.update_sf_intensity_plot(self.Hz)
        elif self.plrz == 'Ez':
            # self.sf_pc.set_array(self.Ez)
            self.plotter.update_sf_intensity_plot(self.Ez)
        else:
            print('What polarization is this?')

        # self.ax.draw_artist(self.sf_pc)
        self.an_t.set_text('time step : {:d}  '.format(self.t))

        # Re-plot the time step annotation window
        self.ax.draw_artist(self.an_t)
        for phc in self.phcs:
            self.ax.draw_artist(phc)

        self.fig.canvas.update()

        # blitting
        # self.ax.figure.canvas.blit(self.ax.bbox)
